"""Quasi-dense instance similarity learning model."""

from typing import List, Tuple

import torch

from vist.struct import Boxes2D, Images, InputSample, LossesType, ModelOutput

from .base import BaseModel, BaseModelConfig, build_model
from .detect import BaseTwoStageDetector
from .track.graph import TrackGraphConfig, build_track_graph
from .track.losses import LossConfig, build_loss
from .track.similarity import SimilarityLearningConfig, build_similarity_head
from .track.utils import cosine_similarity, split_key_ref_inputs


class QDGeneralizedRCNNConfig(BaseModelConfig):
    """Config for quasi-dense tracking model."""

    detection: BaseModelConfig
    similarity: SimilarityLearningConfig
    track_graph: TrackGraphConfig
    losses: List[LossConfig]
    softmax_temp: float = -1.0


class QDGeneralizedRCNN(BaseModel):
    """Generalized R-CNN for quasi-dense instance similarity learning."""

    def __init__(self, cfg: BaseModelConfig) -> None:
        """Init."""
        super().__init__()
        self.cfg = QDGeneralizedRCNNConfig(**cfg.dict())
        self.detector = build_model(self.cfg.detection)
        assert isinstance(self.detector, BaseTwoStageDetector)
        self.similarity_head = build_similarity_head(self.cfg.similarity)
        self.track_graph = build_track_graph(self.cfg.track_graph)
        self.track_loss = build_loss(self.cfg.losses[0])
        self.track_loss_aux = build_loss(self.cfg.losses[1])

    @property
    def device(self) -> torch.device:
        """Get device where input should be moved to."""
        return self.detector.device

    def prepare_targets(
        self,
        key_inputs: List[InputSample],
        ref_inputs: List[List[InputSample]],
    ) -> Tuple[List[Boxes2D], List[List[Boxes2D]]]:
        """Prepare targets from key / ref input samples."""
        key_targets = []
        for x in key_inputs:
            assert x.boxes2d is not None
            key_targets.append(x.boxes2d.to(self.device))
        ref_targets = []
        for inputs in ref_inputs:
            ref_target = []
            for x in inputs:
                assert x.boxes2d is not None
                ref_target.append(x.boxes2d.to(self.device))
            ref_targets.append(ref_target)

        return key_targets, ref_targets

    def prepare_images(
        self,
        key_inputs: List[InputSample],
        ref_inputs: List[List[InputSample]],
    ) -> Tuple[Images, List[Images]]:
        """Prepare images from key / ref input samples."""
        key_images = self.detector.preprocess_image(key_inputs)
        ref_images = [
            self.detector.preprocess_image(inp) for inp in ref_inputs
        ]
        return key_images, ref_images

    def forward_train(
        self, batch_inputs: List[List[InputSample]]
    ) -> LossesType:
        """Forward function for training."""
        # split into key / ref pairs NxM input --> key: N, ref: Nx(M-1)
        key_inputs, ref_inputs = split_key_ref_inputs(batch_inputs)

        # group by ref views by sequence: Nx(M-1) --> (M-1)xN
        ref_inputs = [
            [ref_inputs[j][i] for j in range(len(ref_inputs))]
            for i in range(len(ref_inputs[0]))
        ]

        key_images, ref_images = self.prepare_images(key_inputs, ref_inputs)
        key_targets, ref_targets = self.prepare_targets(key_inputs, ref_inputs)

        # from vist.vis.image import imshow_bboxes
        # for batch_i, key_inp in enumerate(key_inputs):
        #     imshow_bboxes(key_inp.image.tensor[0], key_targets[batch_i])
        #     for ref_i, ref_inp in enumerate(ref_inputs):
        #         imshow_bboxes(
        #             ref_inp[batch_i].image.tensor[0],
        #             ref_targets[ref_i][batch_i],
        #         )

        # feature extraction
        key_x = self.detector.extract_features(key_images)
        ref_x = [self.detector.extract_features(img) for img in ref_images]

        # proposal generation
        key_proposals, rpn_losses = self.detector.generate_proposals(
            key_images, key_x, key_targets
        )
        with torch.no_grad():
            ref_proposals = [
                self.detector.generate_proposals(img, x)[0]
                for img, x in zip(ref_images, ref_x)
            ]

        # bbox head
        _, roi_losses = self.detector.generate_detections(
            key_images,
            key_x,
            key_proposals,
            key_targets,
            compute_detections=False,
        )
        det_losses = {**rpn_losses, **roi_losses}

        # from vist.vis.track import imshow_bboxes
        # for ref_imgs, ref_props in zip(ref_images, ref_proposals):
        #     for ref_img, ref_prop in zip(ref_imgs, ref_props):
        #         _, topk_i = torch.topk(ref_prop.boxes[:, -1], 100)
        #         imshow_bboxes(ref_img.tensor[0], ref_prop[topk_i])

        # track head
        key_embeddings, key_track_targets = self.similarity_head(
            key_images,
            key_x,
            key_proposals,
            key_targets,
            filter_negatives=True,
        )
        ref_track_targets, ref_embeddings = [], []
        for inp, x, proposal, target in zip(
            ref_images, ref_x, ref_proposals, ref_targets
        ):
            embeds, targets = self.similarity_head(inp, x, proposal, target)
            ref_embeddings += [embeds]
            ref_track_targets += [targets]

        # from vist.vis.track import visualize_matches
        # for ref_i, ref_inp in enumerate(ref_inputs):
        #     key_imgs = [key_inputs[i].image.tensor[0]
        #                 for i in range(len(key_inputs))]
        #     ref_imgs = [ref_inp[i].image.tensor[0]
        #                 for i in range(len(key_inputs))]
        #     visualize_matches(key_imgs, ref_imgs, key_track_targets,
        #     ref_track_targets[ref_i])

        track_losses = self.tracking_loss(
            key_embeddings,
            key_track_targets,
            ref_embeddings,
            ref_track_targets,
        )
        return {**det_losses, **track_losses}

    def match(
        self,
        key_embeds: Tuple[torch.Tensor],
        ref_embeds: List[Tuple[torch.Tensor]],
    ) -> Tuple[List[List[torch.Tensor]], List[List[torch.Tensor]]]:
        """Match key / ref embeddings based on cosine similarity."""
        # for each reference view
        dists, cos_dists = [], []
        for ref_embed in ref_embeds:
            # for each batch element
            dists_curr, cos_dists_curr = [], []
            for key_embed, ref_embed_ in zip(key_embeds, ref_embed):
                dist = cosine_similarity(
                    key_embed,
                    ref_embed_,
                    normalize=False,
                    temperature=self.cfg.softmax_temp,
                )
                dists_curr.append(dist)
                if self.track_loss_aux is not None:
                    cos_dist = cosine_similarity(key_embed, ref_embed_)
                    cos_dists_curr.append(cos_dist)

            dists.append(dists_curr)
            cos_dists.append(cos_dists_curr)
        return dists, cos_dists

    @staticmethod
    def get_track_targets(
        key_targets: List[Boxes2D], ref_targets: List[List[Boxes2D]]
    ) -> Tuple[List[List[torch.Tensor]], List[List[torch.Tensor]]]:
        """Create tracking target tensors."""
        # for each reference view
        track_targets, track_weights = [], []
        for ref_target in ref_targets:
            # for each batch element
            curr_targets, curr_weights = [], []
            for key_target, ref_target_ in zip(key_targets, ref_target):
                assert (
                    key_target.track_ids is not None
                    and ref_target_.track_ids is not None
                )
                # target shape: len(key_target) x len(ref_target_)
                target = (
                    key_target.track_ids.view(-1, 1)
                    == ref_target_.track_ids.view(1, -1)
                ).int()
                weight = (target.sum(dim=1) > 0).float()
                curr_targets.append(target)
                curr_weights.append(weight)
            track_targets.append(curr_targets)
            track_weights.append(curr_weights)
        return track_targets, track_weights

    def tracking_loss(
        self,
        key_embeddings: Tuple[torch.Tensor],
        key_targets: List[Boxes2D],
        ref_embeddings: List[Tuple[torch.Tensor]],
        ref_targets: List[List[Boxes2D]],
    ) -> LossesType:
        """Calculate losses for tracking.

        Key inputs are of type List[Tensor/Boxes2D] (Lists are length N)
        Ref inputs are of type List[List[Tensor/Boxes2D]] where the lists
        are of length MxN.
        Where M is the number of reference views and N is the
        number of batch elements.
        """
        losses = dict()

        loss_track = torch.tensor(0.0).to(self.device)
        loss_track_aux = torch.tensor(0.0).to(self.device)
        dists, cos_dists = self.match(key_embeddings, ref_embeddings)
        track_targets, track_weights = self.get_track_targets(
            key_targets, ref_targets
        )
        # for each reference view
        for curr_dists, curr_cos_dists, curr_targets, curr_weights in zip(
            dists, cos_dists, track_targets, track_weights
        ):
            # for each batch element
            for _dists, _cos_dists, _targets, _weights in zip(
                curr_dists, curr_cos_dists, curr_targets, curr_weights
            ):
                if all(_dists.shape):
                    loss_track += self.track_loss(
                        _dists,
                        _targets,
                        _weights,
                        avg_factor=_weights.sum() + 1e-5,
                    )
                    if self.track_loss_aux is not None:
                        loss_track_aux += self.track_loss_aux(
                            _cos_dists, _targets
                        )

        num_pairs = len(dists) * len(dists[0])
        losses["track_loss"] = loss_track / num_pairs
        if self.track_loss_aux is not None:
            losses["track_loss_aux"] = loss_track_aux / num_pairs

        return losses

    def forward_test(
        self, batch_inputs: List[InputSample], postprocess: bool = True
    ) -> ModelOutput:
        """Forward function during inference."""
        assert len(batch_inputs) == 1, "Currently only BS=1 supported!"

        # init graph at begin of sequence
        frame_id = batch_inputs[0].metadata.frame_index
        if frame_id == 0:
            self.track_graph.reset()

        # detector
        image = self.detector.preprocess_image(batch_inputs)
        feat = self.detector.extract_features(image)
        proposals, _ = self.detector.generate_proposals(image, feat)
        detections, _ = self.detector.generate_detections(
            image, feat, proposals
        )
        assert detections is not None

        # from vist.vis.image import imshow_bboxes
        # imshow_bboxes(image.tensor[0], detections)

        # similarity head
        embeddings, _ = self.similarity_head(image, feat, detections)
        if postprocess:
            ori_wh = (
                batch_inputs[0].metadata.size.width,  # type: ignore
                batch_inputs[0].metadata.size.height,  # type: ignore
            )
            self.postprocess(ori_wh, image.image_sizes[0], detections[0])

        # associate detections, update graph
        tracks = self.track_graph(detections[0], frame_id, embeddings[0])

        return dict(detect=detections, track=[tracks])



"""RoI heads."""
from .base import (
    BaseSimilarityHead,
    SimilarityLearningConfig,
    build_similarity_head,
)
from .quasi_dense_embedding_head import QDSimilarityHead

__all__ = [
    "BaseSimilarityHead",
    "QDSimilarityHead",
    "build_similarity_head",
    "SimilarityLearningConfig",
]



"""Testcases for dataset build functions."""
import unittest

from detectron2.data import DatasetCatalog, MetadataCatalog
from scalabel.label.typing import Frame

from vist.data.build import get_dataset_frames


class TestBuild(unittest.TestCase):
    """build Testcase class."""

    def test_get_dataset_dicts(self) -> None:
        """Testcase for get_dataset_dicts."""
        names = ["a", "b"]
        cats = ["car"]
        data_dicts = [
            [Frame(**dict(name="1", labels=[dict(id="1", category="car")]))],
            [Frame(**dict(name="1", labels=[dict(id="1", category="car")]))],
        ]

        DatasetCatalog.register(names[0], lambda: data_dicts[0])
        DatasetCatalog.register(names[1], lambda: data_dicts[1])

        for name in names:
            meta = MetadataCatalog.get(name)
            meta.thing_classes = cats
            meta.idx_to_class_mapping = dict(enumerate(cats))
            meta.class_frequencies = {c: 0 for c in cats}

        dataset_dicts = get_dataset_frames(names, True)
        self.assertEqual(dataset_dicts, [d[0] for d in data_dicts])

        names = ["c", "d", "e"]
        data_dicts = [
            [Frame(**dict(name="1", labels=[]))],
            [Frame(**dict(name="1", labels=None))],
            [
                Frame(
                    **dict(
                        name="1",
                        labels=[
                            dict(
                                id="1",
                                category="car",
                                attributes=dict(crowd=False),
                            )
                        ],
                    )
                )
            ],
        ]

        DatasetCatalog.register(names[0], lambda: data_dicts[0])
        DatasetCatalog.register(names[1], lambda: data_dicts[1])
        DatasetCatalog.register(names[2], lambda: data_dicts[2])

        for name in names:
            meta = MetadataCatalog.get(name)
            meta.thing_classes = cats
            meta.idx_to_class_mapping = dict(enumerate(cats))



"""Tracking base class."""

import abc

import torch
from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D


class TrackGraphConfig(BaseModel, extra="allow"):
    """Base config for tracking graph optimization."""

    type: str = Field(...)


class BaseTrackGraph(torch.nn.Module, metaclass=RegistryHolder):  # type: ignore # pylint: disable=line-too-long
    """Base class for tracking graph optimization."""

    def __init__(self, cfg: TrackGraphConfig):
        """Init."""
        super().__init__()
        self.cfg_base = cfg
        self.reset()

    def reset(self) -> None:
        """Reset tracks."""
        self.num_tracks = 0
        self.tracks = dict()  # type: ignore

    @property
    def empty(self) -> bool:
        """Whether track memory is empty."""
        return not self.tracks

    @abc.abstractmethod
    def forward(  # type: ignore
        self, detections: Boxes2D, frame_id: int, *args, **kwargs
    ) -> Boxes2D:
        """Process inputs, match detections with existing tracks."""
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, *args, **kwargs) -> None:  # type: ignore
        """Update track memory using matched detections."""
        raise NotImplementedError


def build_track_graph(cfg: TrackGraphConfig) -> BaseTrackGraph:
    """Build a tracking graph optimizer from config."""
    registry = RegistryHolder.get_registry(BaseTrackGraph)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseTrackGraph)
        return module
    raise NotImplementedError(f"TrackGraph {cfg.type} not found.")



"""Combined Sampler."""
import inspect
from typing import List, Tuple

import torch
from pydantic import validator

from vist.struct import Boxes2D

from ..matchers.base import MatchResult
from ..utils import non_intersection, random_choice
from .base import BaseSampler, SamplerConfig
from .utils import nonzero_tuple, prepare_target


class CombinedSamplerConfig(SamplerConfig):
    """Config for CombinedSampler."""

    pos_strategy: str
    neg_strategy: str
    neg_pos_ub: float = 3.0
    floor_thr: float = -1.0
    floor_fraction: float = 0.0
    num_bins: int = 3

    # Disable some pylint options in validator:
    # https://github.com/samuelcolvin/pydantic/issues/568
    @validator("pos_strategy", check_fields=False)
    def validate_pos_strategy(  # pylint: disable=no-self-argument,no-self-use
        cls, value: str
    ) -> str:
        """Check pos_strategy attribute."""
        if not value in ["instance_balanced", "iou_balanced"]:
            raise ValueError(
                "pos_strategy must be in [instance_balanced, iou_balanced]"
            )
        return value

    @validator("neg_strategy", check_fields=False)
    def validate_neg_strategy(  # pylint: disable=no-self-argument,no-self-use
        cls, value: str
    ) -> str:
        """Check neg_strategy attribute."""
        if not value in ["instance_balanced", "iou_balanced"]:
            raise ValueError(
                "neg_strategy must be in [instance_balanced, iou_balanced]"
            )
        return value


class CombinedSampler(BaseSampler):
    """Combined sampler. Can have different strategies for pos/neg samples."""

    def __init__(self, cfg: SamplerConfig):
        """Init."""
        super().__init__()
        self.cfg = CombinedSamplerConfig(**cfg.dict())
        self.bg_label = 0
        self.pos_strategy = getattr(self, self.cfg.pos_strategy + "_sampling")
        self.pos_args = inspect.signature(self.pos_strategy).parameters.keys()
        self.neg_strategy = getattr(self, self.cfg.neg_strategy + "_sampling")
        self.neg_args = inspect.signature(self.neg_strategy).parameters.keys()

    @staticmethod
    def instance_balanced_sampling(
        idx_tensor: torch.Tensor, assigned_gts: torch.Tensor, sample_size: int
    ) -> torch.Tensor:
        """Sample indices with balancing according to matched GT instance."""
        if idx_tensor.numel() <= sample_size:
            return idx_tensor

        unique_gt_inds = assigned_gts.unique()
        num_gts = len(unique_gt_inds)
        num_per_gt = int(sample_size / float(num_gts))
        sampled_inds = []
        # sample specific amount per gt instance
        for i in unique_gt_inds:
            inds = torch.nonzero(assigned_gts == i, as_tuple=False)
            inds = inds.squeeze(1)
            if len(inds) > num_per_gt:
                inds = random_choice(inds, num_per_gt)
            sampled_inds.append(inds)
        sampled_inds = torch.cat(sampled_inds)

        # deal with edge cases
        if len(sampled_inds) < sample_size:
            num_extra = sample_size - len(sampled_inds)
            extra_inds = non_intersection(idx_tensor, sampled_inds)
            if len(extra_inds) > num_extra:
                extra_inds = random_choice(extra_inds, num_extra)
            sampled_inds = torch.cat([sampled_inds, extra_inds])
        return sampled_inds

    def iou_balanced_sampling(
        self,
        idx_tensor: torch.Tensor,
        assigned_gt_ious: torch.Tensor,
        sample_size: int,
    ) -> torch.Tensor:
        """Sample indices with balancing according to IoU with matched GT."""
        if idx_tensor.numel() <= sample_size:
            return idx_tensor

        # define 'floor' set - set with low iou samples
        if self.cfg.floor_thr >= 0:
            floor_set = idx_tensor[assigned_gt_ious <= self.cfg.floor_thr]
            iou_sampling_set = idx_tensor[
                assigned_gt_ious > self.cfg.floor_thr
            ]
        else:
            floor_set = None
            iou_sampling_set = idx_tensor[
                assigned_gt_ious > self.cfg.floor_thr
            ]

        num_iou_set_samples = int(sample_size * (1 - self.cfg.floor_fraction))
        if len(iou_sampling_set) > num_iou_set_samples:
            if self.cfg.num_bins >= 2:
                iou_sampled_inds = self.sample_within_intervals(
                    idx_tensor, assigned_gt_ious, num_iou_set_samples
                )
            else:
                iou_sampled_inds = random_choice(
                    iou_sampling_set, num_iou_set_samples
                )
        else:
            iou_sampled_inds = iou_sampling_set  # pragma: no cover

        if floor_set is not None:
            num_floor_set_samples = sample_size - len(iou_sampled_inds)
            if len(floor_set) > num_floor_set_samples:
                sampled_floor_inds = random_choice(
                    floor_set, num_floor_set_samples
                )
            else:
                sampled_floor_inds = floor_set  # pragma: no cover
            sampled_inds = torch.cat([sampled_floor_inds, iou_sampled_inds])
        else:
            sampled_inds = iou_sampled_inds

        if len(sampled_inds) < sample_size:  # pragma: no cover
            num_extra = sample_size - len(sampled_inds)
            extra_inds = non_intersection(idx_tensor, sampled_inds)
            if len(extra_inds) > num_extra:
                extra_inds = random_choice(extra_inds, num_extra)
            sampled_inds = torch.cat([sampled_inds, extra_inds])

        return sampled_inds

    def sample(
        self,
        matching: List[MatchResult],
        boxes: List[Boxes2D],
        targets: List[Boxes2D],
    ) -> Tuple[List[Boxes2D], List[Boxes2D]]:
        """Sample boxes according to strategies defined in cfg."""
        pos_sample_size = int(
            self.cfg.batch_size_per_image * self.cfg.positive_fraction
        )
        sampled_boxes, sampled_targets = [], []
        for match, box, target in zip(matching, boxes, targets):
            positive_mask = (match.assigned_labels != -1) & (
                match.assigned_labels != self.bg_label
            )
            negative_mask = match.assigned_labels == self.bg_label

            positive = nonzero_tuple(positive_mask)[0]
            negative = nonzero_tuple(negative_mask)[0]

            num_pos = min(positive.numel(), pos_sample_size)
            num_neg = self.cfg.batch_size_per_image - num_pos

            if self.cfg.neg_pos_ub >= 0:
                neg_upper_bound = int(self.cfg.neg_pos_ub * num_pos)
                num_neg = min(num_neg, neg_upper_bound)

            args = dict(
                idx_tensor=positive,
                assigned_gts=match.assigned_gt_indices.long()[positive_mask],
                assigned_gt_ious=match.assigned_gt_iou[positive_mask],
                sample_size=num_pos,
            )
            pos_idx = self.pos_strategy(**{k: args[k] for k in self.pos_args})

            args = dict(
                idx_tensor=negative,
                assigned_gts=match.assigned_gt_indices.long()[negative_mask],
                assigned_gt_ious=match.assigned_gt_iou[negative_mask],
                sample_size=num_neg,
            )
            neg_idx = self.neg_strategy(**{k: args[k] for k in self.neg_args})

            sampled_idcs = torch.cat([pos_idx, neg_idx], dim=0)
            sampled_boxes.append(box[sampled_idcs])
            sampled_targets.append(
                prepare_target(len(pos_idx), sampled_idcs, target, match)
            )
        return sampled_boxes, sampled_targets

    def sample_within_intervals(
        self,
        idx_tensor: torch.Tensor,
        assigned_gt_ious: torch.Tensor,
        sample_size: int,
    ) -> torch.Tensor:
        """Sample according to N iou intervals where N = num bins."""
        floor_thr = max(self.cfg.floor_thr, 0.0)
        max_iou = assigned_gt_ious.max()
        iou_interval = (max_iou - floor_thr) / self.cfg.num_bins
        per_bin_samples = int(sample_size / self.cfg.num_bins)

        sampled_inds = []
        for i in range(self.cfg.num_bins):
            start_iou = floor_thr + i * iou_interval
            end_iou = floor_thr + (i + 1) * iou_interval
            tmp_set = nonzero_tuple(
                (start_iou <= assigned_gt_ious) & (assigned_gt_ious < end_iou)
            )[0]
            if len(tmp_set) > per_bin_samples:
                tmp_sampled_set = random_choice(
                    idx_tensor[tmp_set], per_bin_samples
                )
            else:
                tmp_sampled_set = idx_tensor[tmp_set]  # pragma: no cover
            sampled_inds.append(tmp_sampled_set)

        sampled_inds = torch.cat(sampled_inds)
        if len(sampled_inds) < sample_size:
            num_extra = sample_size - len(sampled_inds)
            extra_inds = non_intersection(idx_tensor, sampled_inds)
            if len(extra_inds) > num_extra:
                extra_inds = random_choice(extra_inds, num_extra)
            sampled_inds = torch.cat([sampled_inds, extra_inds])

        return sampled_inds



"""Matchers."""
import abc
from typing import List, NamedTuple

import torch
from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D


class MatcherConfig(BaseModel, extra="allow"):
    """Matcher base config."""

    type: str = Field(...)


class MatchResult(NamedTuple):
    """Match result class. Stores expected result tensors.

    assigned_gt_indices: torch.Tensor - Tensor of [0, M) where M = num gt
    assigned_gt_iou: torch.Tensor  - Tensor with IoU to assigned GT
    assigned_labels: torch.Tensor  - Tensor of {0, -1, 1} = {neg, ignore, pos}
    """

    assigned_gt_indices: torch.Tensor
    assigned_gt_iou: torch.Tensor
    assigned_labels: torch.Tensor


class BaseMatcher(metaclass=RegistryHolder):
    """Base class for box / target matchers."""

    @abc.abstractmethod
    def match(
        self, boxes: List[Boxes2D], targets: List[Boxes2D]
    ) -> List[MatchResult]:
        """Match bounding boxes according to their struct."""
        raise NotImplementedError


def build_matcher(cfg: MatcherConfig) -> BaseMatcher:
    """Build a bounding box matcher from config."""
    registry = RegistryHolder.get_registry(BaseMatcher)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseMatcher)
        return module
    raise NotImplementedError(f"Matcher {cfg.type} not found.")



"""VisT Input data structures."""

import itertools
from typing import Dict, List, Optional, Tuple, Union

import torch
from scalabel.label.typing import Frame

from .labels import Boxes2D, Boxes3D
from .structures import DataInstance


class Intrinsics(DataInstance):
    """Data structure for intrinsic calibration."""

    def __init__(self, tensor: torch.Tensor):
        """Init Intrinsics class.

        Args:
            tensor (torch.Tensor): (N, 3, 3) or (3, 3)
        """
        assert 2 <= len(tensor.shape) <= 3
        if len(tensor.shape) == 3:
            self.batched = True
        else:
            self.batched = False
        self.tensor = tensor

    def to(self, device: torch.device) -> "Intrinsics":
        """Put images on device."""
        cast_tensor = self.tensor.to(device)
        return Intrinsics(cast_tensor)

    @property
    def device(self) -> torch.device:
        """Returns current device."""
        return self.tensor.device

    @classmethod
    def cat(
        cls,
        instances: List["Intrinsics"],
        device: Optional[torch.device] = None,
    ) -> "Intrinsics":
        """Concatenate N Intrinsics objects."""
        tensors = []
        if device is None:
            device = instances[0].tensor.device
        for inst in instances:
            tensor = inst.tensor.to(device)
            if not inst.batched:
                tensor = tensor.unsqueeze(0)
            tensors.append(tensor)
        return Intrinsics(torch.cat(tensors, 0))


class Extrinsics(DataInstance):
    """Data structure for extrinsic calibration."""

    def __init__(self, tensor: torch.Tensor):
        """Init Extrinsics class.

        Args:
            tensor (torch.Tensor): (N, 4, 4) or (4, 4)
        """
        assert 2 <= len(tensor.shape) <= 3
        if len(tensor.shape) == 3:
            self.batched = True
        else:
            self.batched = False
        self.tensor = tensor

    def to(self, device: torch.device) -> "Extrinsics":
        """Put images on device."""
        cast_tensor = self.tensor.to(device)
        return Extrinsics(cast_tensor)

    @property
    def device(self) -> torch.device:
        """Returns current device."""
        return self.tensor.device

    @classmethod
    def cat(
        cls,
        instances: List["Extrinsics"],
        device: Optional[torch.device] = None,
    ) -> "Extrinsics":
        """Concatenate N Extrinsics objects."""
        tensors = []
        if device is None:
            device = instances[0].tensor.device
        for inst in instances:
            tensor = inst.tensor.to(device)
            if not inst.batched:
                tensor = tensor.unsqueeze(0)
            tensors.append(tensor)
        return Extrinsics(torch.cat(tensors, 0))


class Images(DataInstance):
    """Data structure for saving images."""

    stride: int = 32

    def __init__(
        self, tensor: torch.Tensor, image_sizes: List[Tuple[int, int]]
    ):
        """Init Images class.

        Args:
            tensor (torch.Tensor): shape (N, C_1, ..., C_K, H, W) where K >= 1.
            image_sizes (list[tuple[int, int]]): Each tuple is (w, h). It can
                be smaller than (w, h) due to padding.
        """
        assert len(tensor.shape) > 3
        self.tensor = tensor
        self.image_sizes = image_sizes

    def __len__(self) -> int:
        """Return number of images."""
        return len(self.image_sizes)

    def __getitem__(self, idx: int) -> "Images":
        """Access the individual image in its original size.

        Args:
            idx: int or slice

        Returns:
            Tensor: an image of shape (C_1, ..., C_K, H, W)
            where K >= 1
        """
        size = self.image_sizes[idx]
        return Images(
            self.tensor[idx : idx + 1, ..., : size[1], : size[0]], [size]
        )

    def to(self, device: torch.device) -> "Images":
        """Put images on device."""
        cast_tensor = self.tensor.to(device)
        return Images(cast_tensor, self.image_sizes)

    @classmethod
    def cat(
        cls, instances: List["Images"], device: Optional[torch.device] = None
    ) -> "Images":
        """Concatenate N Images objects."""
        assert isinstance(instances, (list, tuple))
        assert len(instances) > 0
        assert all((isinstance(inst, Images) for inst in instances))
        max_hw = (
            max([im.tensor.shape[-2] for im in instances]),
            max([im.tensor.shape[-1] for im in instances]),
        )
        lens = [len(x) for x in instances]

        # ensure divisibility by stride
        pad = lambda x: (x + (cls.stride - 1)) // cls.stride * cls.stride
        max_hw = (pad(x) for x in max_hw)  # type: ignore

        batch_shape = (
            [sum(lens)] + list(instances[0].tensor.shape[1:-2]) + list(max_hw)
        )
        if device is None:
            device = instances[0].tensor.device
        pad_imgs = torch.zeros(batch_shape, device=device)
        cum_len = 0
        for img, cur_len in zip(instances, lens):
            pad_imgs[
                cum_len : cum_len + cur_len,
                ...,
                : img.tensor.shape[-2],
                : img.tensor.shape[-1],
            ].copy_(img.tensor)
            cum_len += cur_len

        all_sizes = list(
            itertools.chain.from_iterable([x.image_sizes for x in instances])
        )
        return Images(pad_imgs.contiguous(), all_sizes)

    @property
    def device(self) -> torch.device:
        """Returns current device."""
        return self.tensor.device


class InputSample:
    """Container holding varying types of DataInstances and Frame metadata."""

    def __init__(
        self,
        metadata: Frame,
        image: Images,
        boxes2d: Optional[Boxes2D] = None,
        boxes3d: Optional[Boxes3D] = None,
        intrinsics: Optional[Intrinsics] = None,
        extrinsics: Optional[Extrinsics] = None,
        **kwargs: DataInstance,
    ) -> None:
        """Init."""
        self.metadata = metadata
        self.image = image
        if boxes2d is None:
            boxes2d = Boxes2D(
                torch.empty(0, 5), torch.empty(0), torch.empty(0)
            )
        self.boxes2d = boxes2d

        if boxes3d is None:
            boxes3d = Boxes3D(
                torch.empty(0, 8), torch.empty(0), torch.empty(0)
            )
        self.boxes3d = boxes3d

        if intrinsics is None:
            intrinsics = Intrinsics(torch.eye(3))
        self.intrinsics = intrinsics

        if extrinsics is None:
            extrinsics = Extrinsics(torch.eye(4))
        self.extrinsics = extrinsics

        self.attributes = dict()  # type: Dict[str, DataInstance]
        for k, v in kwargs.items():
            self.attributes[k] = v

    def get(self, key: str) -> Union[Frame, DataInstance]:
        """Get attribute by key."""
        if key == "metadata":
            return self.metadata
        if key == "image":
            return self.image
        if key == "boxes2d":
            return self.boxes2d
        if key == "boxes3d":
            return self.boxes3d
        if key == "intrinsics":
            return self.intrinsics
        if key == "extrinsics":
            return self.extrinsics
        if key in self.attributes:
            return self.attributes[key]
        raise AttributeError(f"Attribute {key} not found!")

    def dict(self) -> Dict[str, Union[Frame, DataInstance]]:
        """Return InputSample object as dict."""
        obj_dict = {
            "metadata": self.metadata,
            "image": self.image,
            "boxes2d": self.boxes2d,
            "boxes3d": self.boxes3d,
            "intrinsics": self.intrinsics,
            "extrinsics": self.extrinsics,
        }  # type: Dict[str, Union[Frame, DataInstance]]
        obj_dict.update(self.attributes)
        return obj_dict



"""Test cases for models engine config."""
import unittest
from argparse import Namespace

from vist import config
from vist.common.utils import timeit
from vist.data.datasets.base import BaseDatasetConfig
from vist.engine.utils import _register, to_detectron2
from vist.unittest.utils import get_test_file


class TestConfig(unittest.TestCase):
    """Test cases for vist models config."""

    @timeit  # type: ignore
    def test_register(self) -> None:
        """Testcase for register function."""
        datasets = [
            BaseDatasetConfig(
                **dict(
                    name="example",
                    type="MOTChallenge",
                    data_root="/path/to/data",
                    annotations="/path/to/annotations",
                )
            )
        ]
        names = _register(datasets)
        self.assertEqual(names, ["example"])

    def test_to_detectron2(self) -> None:
        """Testcase for detectron2 config conversion."""
        test_file = get_test_file("detect/faster_rcnn_d2.toml")
        args = Namespace(config=test_file, device="cuda")
        cfg = config.parse_config(args)
        det2cfg = to_detectron2(cfg)
        self.assertEqual(
            det2cfg.SOLVER.IMS_PER_BATCH, cfg.solver.images_per_gpu
        )
        self.assertEqual(
            det2cfg.DATALOADER.NUM_WORKERS, cfg.dataloader.workers_per_gpu
        )
        self.assertEqual(det2cfg.DATASETS.TRAIN, ["bdd100k_det_sample_train"])
        self.assertEqual(det2cfg.DATASETS.TEST, ["bdd100k_det_sample_val"])



"""Utilities for mmdet wrapper."""

import os
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import requests
import torch
from mmcv import Config as MMConfig

from vist.struct import Boxes2D, Images, LossesType

from ..base import BaseModelConfig

MMDetMetaData = Dict[str, Union[Tuple[int, int, int], bool, float]]


class MMTwoStageDetectorConfig(BaseModelConfig):
    """Config for mmdetection two stage models."""

    model_base: str
    model_kwargs: Optional[Dict[str, Union[bool, float, str, List[float]]]]
    num_classes: Optional[int]
    pixel_mean: Tuple[float, float, float]
    pixel_std: Tuple[float, float, float]
    backbone_output_names: Optional[List[str]]
    weights: Optional[str]


def get_img_metas(images: Images) -> List[MMDetMetaData]:
    """Create image metadata in mmdetection format."""
    img_metas = []
    _, c, padh, padw = images.tensor.shape  # type: Tuple[int, int, int, int]
    for i in range(len(images)):
        meta = dict()  # type: MMDetMetaData
        w, h = images.image_sizes[i]
        meta["img_shape"] = meta["ori_shape"] = (h, w, c)
        meta["scale_factor"] = 1.0
        meta["flip"] = False
        meta["pad_shape"] = (padh, padw, c)
        img_metas.append(meta)

    return img_metas


def proposals_from_mmdet(proposals: List[torch.Tensor]) -> List[Boxes2D]:
    """Convert mmdetection proposals to VisT format."""
    proposals_boxes2d = []
    for proposal in proposals:
        proposals_boxes2d.append(Boxes2D(proposal))
    return proposals_boxes2d


def proposals_to_mmdet(proposals: List[Boxes2D]) -> List[torch.Tensor]:
    """Convert VisT format proposals to mmdetection."""
    proposal_tensors = []
    for proposal in proposals:
        proposal_tensors.append(proposal.boxes)
    return proposal_tensors


def detections_from_mmdet(
    bboxes: List[torch.Tensor], labels: List[torch.Tensor]
) -> List[Boxes2D]:
    """Convert mmdetection detections to VisT format."""
    detections_boxes2d = []
    for bbox, label in zip(bboxes, labels):
        if not label.device == bbox.device:
            label = label.to(bbox.device)  # pragma: no cover
        detections_boxes2d.append(Boxes2D(bbox, label))
    return detections_boxes2d


def results_from_mmdet(
    detections: List[torch.Tensor], device: torch.device
) -> List[Boxes2D]:
    """Convert mmdetection bbox_results to VisT format."""
    detections_boxes2d = []
    for detection in detections:
        bboxes = torch.from_numpy(np.vstack(detection)).to(device)  # Nx5
        labels = [
            torch.full((bbox.shape[0],), i, dtype=torch.int32, device=device)
            for i, bbox in enumerate(detection)
        ]
        labels = torch.cat(labels)
        detections_boxes2d.append(Boxes2D(bboxes, labels))
    return detections_boxes2d


def targets_to_mmdet(
    targets: List[Boxes2D],
) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
    """Convert VisT targets to mmdetection compatible format."""
    gt_bboxes = [t.boxes for t in targets]

    gt_labels = [t.class_ids for t in targets]
    return gt_bboxes, gt_labels


def load_config_from_mmdet(url: str) -> str:
    """Get config from mmdetection GitHub repository."""
    full_url = (
        "https://raw.githubusercontent.com/"
        "open-mmlab/mmdetection/master/configs/" + url
    )
    response = requests.get(full_url)
    assert (
        response.status_code == 200
    ), f"Request to {full_url} failed with code {response.status_code}!"
    return response.text


def get_mmdet_config(config: MMTwoStageDetectorConfig) -> MMConfig:
    """Convert a Detector config to a mmdet readable config."""
    if os.path.exists(config.model_base):
        cfg = MMConfig.fromfile(config.model_base)
        if cfg.get("model"):
            cfg = cfg["model"]
    elif config.model_base.startswith("mmdet://"):
        ext = os.path.splitext(config.model_base)[1]
        cfg = MMConfig.fromstring(
            load_config_from_mmdet(config.model_base.strip("mmdet://")), ext
        ).model
    else:
        raise FileNotFoundError(
            f"MMDetection config not found: {config.model_base}"
        )

    if config.num_classes:
        cfg["roi_head"]["bbox_head"]["num_classes"] = config.num_classes

    # add keyword args in config
    if config.model_kwargs:
        for k, v in config.model_kwargs.items():
            attr = cfg
            partial_keys = k.split(".")
            partial_keys, last_key = partial_keys[:-1], partial_keys[-1]
            for part_k in partial_keys:
                attr = attr.get(part_k)
            if attr.get(last_key) is not None:
                attr[last_key] = type(attr.get(last_key))(v)
            else:
                attr[last_key] = v

    return cfg


def _parse_losses(losses: Dict[str, torch.Tensor]) -> LossesType:
    """Parse losses to a scalar tensor."""
    log_vars = dict()
    for name, value in losses.items():
        if "loss" in name:
            if isinstance(value, torch.Tensor):
                log_vars[name] = value.mean()
            elif isinstance(value, list):
                log_vars[name] = sum(_loss.mean() for _loss in value)
            else:
                raise ValueError(f"{name} is not a tensor or list of tensors")

    return log_vars



"""Test cases for VisT data structures."""
import unittest

import torch
from scalabel.label.typing import Frame

from .data import Extrinsics, Images, InputSample, Intrinsics


class TestDataStructures(unittest.TestCase):
    """Test cases VisT data structures."""

    im1 = Images(torch.zeros(2, 1, 128, 128), [(110, 110), (120, 120)])
    im2 = Images(torch.zeros(1, 1, 138, 138), [(130, 130)])

    intr1 = Intrinsics(torch.zeros(3, 3))
    intr2 = Intrinsics(torch.ones(2, 3, 3))

    extr1 = Extrinsics(torch.zeros(4, 4))
    extr2 = Extrinsics(torch.ones(2, 4, 4))

    def test_intrinsics(self) -> None:
        """Testcases for intrinsics class."""
        intrs = Intrinsics.cat([self.intr1, self.intr2])
        self.assertEqual(intrs.tensor.shape, (3, 3, 3))
        ims = intrs.to(torch.device("cpu"))
        self.assertEqual(ims.device, torch.device("cpu"))

    def test_extrinsics(self) -> None:
        """Testcases for extrinsics class."""
        extrs = Extrinsics.cat([self.extr1, self.extr2])
        self.assertEqual(extrs.tensor.shape, (3, 4, 4))
        ims = extrs.to(torch.device("cpu"))
        self.assertEqual(ims.device, torch.device("cpu"))

    def test_images(self) -> None:
        """Testcases for images class."""
        Images.stride = 16
        ims = Images.cat([self.im1, self.im2])
        self.assertEqual(ims.tensor.shape, (3, 1, 144, 144))
        self.assertEqual(ims[0].tensor.shape, (1, 1, 110, 110))
        ims = ims.to(torch.device("cpu"))
        self.assertEqual(ims.device, torch.device("cpu"))

    def test_inputsample(self) -> None:
        """Testcases for InputSample class."""
        attributes = [
            "metadata",
            "image",
            "boxes2d",
            "boxes3d",
            "intrinsics",
            "extrinsics",
            "image2",
        ]
        sample = InputSample(Frame(name="f1"), self.im1, image2=self.im2)
        self.assertEqual(sample.metadata.name, "f1")
        meta = sample.get("metadata")
        assert isinstance(meta, Frame)
        self.assertEqual(meta.name, "f1")  # pylint: disable=no-member
        self.assertEqual(sample.image.tensor.shape, self.im1.tensor.shape)
        image2 = sample.get("image2")
        assert isinstance(image2, Images)
        self.assertEqual(image2.tensor.shape, self.im2.tensor.shape)
        self.assertEqual(list(sample.dict().keys()), attributes)
        for attr in attributes:
            sample.get(attr)



"""Matchers package."""
from .base import BaseMatcher, MatcherConfig, MatchResult, build_matcher
from .max_iou import MaxIoUMatcher

__all__ = [
    "BaseMatcher",
    "MaxIoUMatcher",
    "build_matcher",
    "MatcherConfig",
    "MatchResult",
]



"""Test cases for quasi dense tracking graph construction."""
import unittest

import torch

from vist.common.bbox.utils import compute_iou
from vist.struct import Boxes2D
from vist.unittest.utils import generate_dets

from .base import TrackGraphConfig
from .quasi_dense import QDTrackGraph


class TestQDTrackGraph(unittest.TestCase):
    """Test cases for quasi-dense tracking graph construction."""

    def test_get_tracks(self) -> None:
        """Testcase for get tracks method."""
        tracker = QDTrackGraph(
            TrackGraphConfig(type="qdtrack", keep_in_memory=3)
        )

        h, w, num_dets = 128, 128, 64
        detections = generate_dets(h, w, num_dets)
        embeddings = torch.rand(num_dets, 128)

        for i in range(num_dets):
            tracker.create_track(i, detections[i], embeddings[i], 0)

        boxes, embeds = tracker.get_tracks(torch.device("cpu"), frame_id=0)
        self.assertTrue(len(boxes) == len(embeds) == num_dets)

        for i in range(num_dets // 2):
            tracker.update_track(i, detections[i], embeddings[i], 1)

        boxes, embeds = tracker.get_tracks(torch.device("cpu"), frame_id=1)
        self.assertTrue(len(boxes) == len(embeds) == num_dets // 2)

        boxes, embeds = tracker.get_tracks(torch.device("cpu"), frame_id=2)
        self.assertTrue(len(boxes) == len(embeds) == 0)

    def test_track(self) -> None:
        """Testcase for tracking function."""
        tracker = QDTrackGraph(
            TrackGraphConfig(type="qdtrack", keep_in_memory=3)
        )

        h, w, num_dets = 128, 128, 64
        detections = generate_dets(h, w, num_dets)
        embeddings = torch.rand(num_dets, 128)

        # feed same detections & embeddings --> should be matched to self
        result_t0 = tracker(detections, 0, embeddings)
        result_t1 = tracker(detections, 1, embeddings)
        result_t2 = tracker(detections, 2, embeddings)

        empty_det, empty_emb = Boxes2D(torch.empty(0, 5)), torch.empty(0, 128)
        for i in range(tracker.cfg.keep_in_memory + 1):
            result_final = tracker(empty_det, 3 + i, empty_emb)

        self.assertTrue(tracker.empty)
        self.assertEqual(len(result_final), 0)

        # check if matching is correct
        for t0, t1, t2 in zip(
            result_t0.track_ids, result_t1.track_ids, result_t2.track_ids
        ):
            self.assertTrue(t0 == t1 == t2)

        # check if all tracks have scores >= threshold
        for res in [result_t0, result_t1, result_t2]:
            self.assertTrue(
                (res.boxes[:, -1] >= tracker.cfg.obj_score_thr).all()
            )

            # check if tracks do not overlap too much
            ious = compute_iou(res, res) - torch.eye(len(res.boxes))
            self.assertTrue((ious <= tracker.cfg.nms_class_iou_thr).all())



"""data utils."""
import itertools
import logging
import sys
from collections import defaultdict
from io import BytesIO
from typing import Any, Dict, List, Optional, Union

import kornia
import numpy as np
import torch
from detectron2.utils.logger import log_first_n
from fvcore.common.timer import Timer
from PIL import Image
from scalabel.label.typing import Frame
from scalabel.label.utils import check_crowd, check_ignored
from tabulate import tabulate
from termcolor import colored

from vist.struct import NDArrayUI8

D2BoxType = Dict[str, Union[bool, float, str]]
logger = logging.getLogger(__name__)


def transform_bbox(
    trans_mat: torch.Tensor, boxes: torch.Tensor
) -> torch.Tensor:
    """Apply trans_mat (3, 3) / (B, 3, 3)  to (N, 4) / (B, N, 4) xyxy boxes."""
    assert len(trans_mat.shape) == len(
        boxes.shape
    ), "trans_mat and boxes must have same number of dimensions!"
    x1y1 = boxes[..., :2]
    x2y2 = boxes[..., 2:]
    if len(boxes.shape) == 2:
        x1y1 = x1y1.unsqueeze(0)
        x2y2 = x2y2.unsqueeze(0)
        trans_mat = trans_mat.unsqueeze(0)

    x1y1 = kornia.transform_points(trans_mat, x1y1)
    x2y2 = kornia.transform_points(trans_mat, x2y2)

    x1x2 = torch.stack((x1y1[..., 0], x2y2[..., 0]), -1)
    y1y2 = torch.stack((x1y1[..., 1], x2y2[..., 1]), -1)
    transformed_boxes = torch.stack(
        (
            x1x2.min(dim=-1)[0],
            y1y2.min(dim=-1)[0],
            x1x2.max(dim=-1)[0],
            y1y2.max(dim=-1)[0],
        ),
        -1,
    )

    if len(boxes.shape) == 2:
        transformed_boxes.squeeze(0)
    return transformed_boxes


def identity_batch_collator(  # type: ignore
    batch: List[List[Dict[str, Any]]]
) -> List[List[Dict[str, Any]]]:
    """Identity function batch collator."""
    return batch


def im_decode(im_bytes: bytes, mode: str = "BGR") -> NDArrayUI8:
    """Decode to image (numpy array, BGR) from bytes."""
    pil_img = Image.open(BytesIO(bytearray(im_bytes)))
    if mode == "BGR":
        np_img = np.array(pil_img)[..., [2, 1, 0]]  # type: NDArrayUI8
    elif mode == "RGB":
        np_img = np.array(pil_img)
    else:
        raise NotImplementedError(f"{mode} not supported for image decoding!")
    return np_img


def instance_ids_to_global(
    frames: List[Frame], local_instance_ids: Dict[str, List[str]]
) -> None:
    """Use local (per video) instance ids to produce global ones."""
    video_names = list(local_instance_ids.keys())
    for frame_id, ann in enumerate(frames):
        if ann.labels is not None:
            for label in ann.labels:
                assert label.attributes is not None
                if not check_crowd(label) and not check_ignored(label):
                    video_name = (
                        ann.video_name
                        if ann.video_name is not None
                        else "no-video-" + str(frame_id)
                    )
                    sum_previous_vids = sum(
                        (
                            len(local_instance_ids[v])
                            for v in video_names[
                                : video_names.index(video_name)
                            ]
                        )
                    )
                    label.attributes[
                        "instance_id"
                    ] = sum_previous_vids + local_instance_ids[
                        video_name
                    ].index(
                        label.id
                    )


def prepare_labels(
    cat_name2id: Dict[str, int],
    frames: List[Frame],
    global_instance_ids: bool = False,
) -> Dict[str, int]:
    """Add category id and instance id to labels, return class frequencies."""
    timer = Timer()
    instance_ids = defaultdict(list)  # type: Dict[str, List[str]]
    frequencies = {cat: 0 for cat in cat_name2id}
    for frame_id, ann in enumerate(frames):
        if ann.labels is not None:
            for label in ann.labels:
                attr = dict()  # type: Dict[str, Union[bool, int, float, str]]
                if label.attributes is not None:
                    attr = label.attributes

                if not check_crowd(label) and not check_ignored(label):
                    assert label.category is not None
                    frequencies[label.category] += 1
                    attr["category_id"] = cat_name2id[label.category]

                    video_name = (
                        ann.video_name
                        if ann.video_name is not None
                        else "no-video-" + str(frame_id)
                    )
                    if label.id not in instance_ids[video_name]:
                        instance_ids[video_name].append(label.id)
                    attr["instance_id"] = instance_ids[video_name].index(
                        label.id
                    )

                label.attributes = attr

    if global_instance_ids:
        instance_ids_to_global(frames, instance_ids)

    logger.info(
        "Preprocessing %s labels takes %s seconds.",
        len(frames),
        "{:.2f}".format(timer.seconds()),
    )
    return frequencies


def str_decode(str_bytes: bytes, encoding: Optional[str] = None) -> str:
    """Decode to string from bytes."""
    if encoding is None:
        encoding = sys.getdefaultencoding()
    return str_bytes.decode(encoding)


def discard_labels_outside_set(
    dataset: List[Frame], class_set: List[str]
) -> None:
    """Discard labels outside given set of classes."""
    for frame in dataset:
        remove_anns = []
        if frame.labels is not None:
            for i, ann in enumerate(frame.labels):
                if not ann.category in class_set:
                    remove_anns.append(i)
            for i in reversed(remove_anns):
                frame.labels.pop(i)


def print_class_histogram(class_frequencies: Dict[str, int]) -> None:
    """Prints out given class frequencies."""
    class_names = list(class_frequencies.keys())
    frequencies = list(class_frequencies.values())
    num_classes = len(class_names)

    n_cols = min(6, len(class_names) * 2)

    def short_name(name: str) -> str:
        """Make long class names shorter."""
        if len(name) > 13:
            return name[:11] + ".."  # pragma: no cover
        return name

    data = list(
        itertools.chain(
            *[
                [short_name(class_names[i]), int(v)]
                for i, v in enumerate(frequencies)
            ]
        )
    )
    total_num_instances = sum(data[1::2])
    data.extend([None] * (n_cols - (len(data) % n_cols)))
    if num_classes > 1:
        data.extend(["total", total_num_instances])
    data = itertools.zip_longest(*[data[i::n_cols] for i in range(n_cols)])  # type: ignore # pylint: disable=line-too-long
    table = tabulate(
        data,  # type: ignore
        headers=["category", "#instances"] * (n_cols // 2),
        tablefmt="pipe",
        numalign="left",
        stralign="center",
    )
    log_first_n(
        logging.INFO,
        "Distribution of instances among all {} categories:\n".format(
            num_classes
        )
        + colored(table, "cyan"),
        key="message",
    )



"""Default argument parser for vist."""
import argparse

from .config import Launch


def default_argument_parser() -> argparse.ArgumentParser:
    """Create a parser with common vist arguments."""
    schema = Launch.schema()
    parser = argparse.ArgumentParser(description=schema["description"])
    parser.add_argument(
        "action",
        type=str,
        choices=["train", "test", "predict"],
        help="Action to execute",
    )
    parser.add_argument(
        "--config", default="", metavar="FILE", help="path to config file"
    )

    parser.add_argument(
        "--cfg-options",
        default="",
        help="additional config parameters in format key=value separated by "
        "commas, e.g. dataloader.workers_per_gpu=1,solver.base_lr=0.01",
    )

    for key, val in schema["properties"].items():
        if Launch.__fields__[key].type_ == bool:
            if not val["default"]:
                parser.add_argument("--" + key, action="store_true")
            else:
                parser.add_argument(
                    "--" + key, action="store_false"
                )  # pragma: no cover
        else:
            parser.add_argument(
                "--" + key,
                default=None,
                type=Launch.__fields__[key].type_,
            )
    return parser



"""Embedding distance loss."""
from typing import Optional, Tuple

import torch

from vist.common.bbox.utils import random_choice

from .base import BaseLoss, LossConfig
from .utils import weight_reduce_loss


class EmbeddingDistanceLossConfig(LossConfig):
    """Config for embedding distance loss."""

    neg_pos_ub: float = -1.0
    pos_margin: float = -1.0
    neg_margin: float = -1.0
    hard_mining: bool = False


class EmbeddingDistanceLoss(BaseLoss):
    """Embedding distance loss."""

    def __init__(self, cfg: LossConfig):
        """Init."""
        super().__init__()
        self.cfg = EmbeddingDistanceLossConfig(**cfg.dict())

    def forward(  # type: ignore # pylint: disable=arguments-differ
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        weight: Optional[torch.Tensor] = None,
        reduction_override: Optional[str] = None,
    ) -> torch.Tensor:
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction.
            target (torch.Tensor): The learning target of the prediction.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Defaults to None.

        Returns:
            loss_bbox (torch.Tensor): embedding distance loss.
        """
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = (
            reduction_override
            if reduction_override is not None
            else self.cfg.reduction
        )
        pred, weight, avg_factor = self.update_weight(pred, target, weight)
        loss = self.cfg.loss_weight * l2_loss(
            pred, target, weight, reduction=reduction, avg_factor=avg_factor
        )
        return loss

    def update_weight(
        self, pred: torch.Tensor, target: torch.Tensor, weight: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, float]:
        """Update element-wise loss weights.

        Exclude negatives according to maximum fraction of samples and/or
        hard negative mining.
        """
        if weight is None:
            weight = target.new_ones(target.size())
        invalid_inds = weight <= 0
        target[invalid_inds] = -1
        pos_inds = target == 1
        neg_inds = target == 0

        if self.cfg.pos_margin > 0:
            pred[pos_inds] -= self.cfg.pos_margin
        if self.cfg.neg_margin > 0:
            pred[neg_inds] -= self.cfg.neg_margin
        pred = torch.clamp(pred, min=0, max=1)

        num_pos = max(1, int((target == 1).sum()))
        num_neg = int((target == 0).sum())
        if self.cfg.neg_pos_ub > 0 and num_neg / num_pos > self.cfg.neg_pos_ub:
            num_neg = int(num_pos * self.cfg.neg_pos_ub)
            neg_idx = torch.nonzero(target == 0, as_tuple=False)

            if self.cfg.hard_mining:
                costs = l2_loss(pred, target, reduction="none")[
                    neg_idx[:, 0], neg_idx[:, 1]
                ].detach()
                neg_idx = neg_idx[costs.topk(num_neg)[1], :]
            else:
                neg_idx = random_choice(neg_idx, num_neg)

            new_neg_inds = neg_inds.new_zeros(neg_inds.size()).bool()
            new_neg_inds[neg_idx[:, 0], neg_idx[:, 1]] = True

            invalid_neg_inds = torch.logical_xor(neg_inds, new_neg_inds)
            weight[invalid_neg_inds] = 0

        avg_factor = (weight > 0).sum()
        return pred, weight, avg_factor


def l2_loss(
    pred: torch.Tensor,
    target: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    reduction: str = "mean",
    avg_factor: Optional[float] = None,
) -> torch.Tensor:
    """L2 loss."""
    assert pred.size() == target.size() and target.numel() > 0
    loss = torch.abs(pred - target) ** 2
    if weight is not None:
        weight = weight.float()
    return weight_reduce_loss(loss, weight, reduction, avg_factor)



"""Match predictions and targets according to maximum 2D IoU."""
from typing import List

import torch
from detectron2.modeling.matcher import Matcher as D2Matcher

from vist.common.bbox.utils import compute_iou
from vist.struct import Boxes2D

from .base import BaseMatcher, MatcherConfig, MatchResult


class MaxIoUMatcherConfig(MatcherConfig):
    """MaxIoUMatcher config."""

    thresholds: List[float]
    labels: List[int]
    allow_low_quality_matches: bool


class MaxIoUMatcher(BaseMatcher):
    """MaxIoUMatcher class. Based on Matcher in detectron2."""

    def __init__(self, cfg: MatcherConfig):
        """Init."""
        super().__init__()
        self.cfg = MaxIoUMatcherConfig(**cfg.dict())

        self.matcher = D2Matcher(
            thresholds=self.cfg.thresholds,
            labels=self.cfg.labels,
            allow_low_quality_matches=self.cfg.allow_low_quality_matches,
        )

    def match(
        self, boxes: List[Boxes2D], targets: List[Boxes2D]
    ) -> List[MatchResult]:
        """Match all boxes to targets based on maximum IoU."""
        result = []
        for b, t in zip(boxes, targets):
            if len(t) == 0:
                matches = torch.zeros(len(b)).to(b.device)
                match_labels = torch.zeros(len(b)).to(b.device)
                match_iou = torch.zeros(len(b)).to(b.device)
            else:
                # M x N matrix, where M = num gt, N = num proposals
                match_quality_matrix = compute_iou(t, b)

                # matches N x 1 = index of assigned gt i.e.  range [0, M)
                # match_labels N x 1, 0 = negative, -1 = ignore, 1 = positive
                matches, match_labels = self.matcher(match_quality_matrix)
                match_iou = match_quality_matrix[
                    matches, torch.arange(0, len(b)).to(b.device)
                ]

            result.append(
                MatchResult(
                    assigned_gt_indices=matches,
                    assigned_labels=match_labels,
                    assigned_gt_iou=match_iou,
                )
            )
        return result



"""Test cases for VisT annotation data structures."""
import unittest

import torch

from vist.struct import Boxes2D, Boxes3D
from vist.unittest.utils import generate_dets, generate_dets3d


class TestBoxes2D(unittest.TestCase):
    """Test cases VisT Boxes2D."""

    def test_scalabel(self) -> None:
        """Testcase for conversion to / from scalabel."""
        h, w, num_dets = 128, 128, 10
        detections = generate_dets(h, w, num_dets, track_ids=True)
        idx_to_class = {0: "car"}
        class_to_idx = {"car": 0}
        scalabel_dets = detections.to_scalabel(idx_to_class)

        detections_new = Boxes2D.from_scalabel(scalabel_dets, class_to_idx)

        scalabel_dets[0].box2d = None
        dets_with_none = Boxes2D.from_scalabel(scalabel_dets, class_to_idx)
        self.assertTrue(
            torch.isclose(
                dets_with_none.boxes[0], detections_new.boxes[1]
            ).all()
        )

        for det, det_new in zip(detections, detections_new):  # type: ignore
            self.assertTrue(torch.isclose(det.boxes, det_new.boxes).all())
            self.assertTrue(
                torch.isclose(det.class_ids.long(), det_new.class_ids).all()
            )
            self.assertTrue(
                torch.isclose(det.track_ids.long(), det_new.track_ids).all()
            )

        detections_new.track_ids = None
        dets_without_tracks = detections_new.to_scalabel(idx_to_class)
        self.assertTrue(
            all(
                (str(i) == det.id for i, det in enumerate(dets_without_tracks))
            )
        )

        detections.boxes = detections.boxes[:, :-1]
        scalabel_dets_no_score = detections.to_scalabel(idx_to_class)
        self.assertTrue(all(d.score is None for d in scalabel_dets_no_score))

    def test_clone(self) -> None:
        """Testcase for cloning a Boxes2D object."""
        h, w, num_dets = 128, 128, 10
        detections = generate_dets(h, w, num_dets, track_ids=True)
        detections_new = detections.clone()

        for det, det_new in zip(detections, detections_new):  # type: ignore
            self.assertTrue(torch.isclose(det.boxes, det_new.boxes).all())
            self.assertTrue(
                torch.isclose(det.class_ids, det_new.class_ids).all()
            )
            self.assertTrue(
                torch.isclose(det.track_ids, det_new.track_ids).all()
            )

    def test_cat(self) -> None:
        """Testcase for concatenating a list of Boxes2D objects."""
        h, w, num_dets = 128, 128, 10
        det = generate_dets(h, w, num_dets, track_ids=True)
        det_new = Boxes2D.cat([det, det])

        self.assertTrue(
            torch.isclose(
                torch.cat([det.boxes, det.boxes]), det_new.boxes
            ).all()
        )

        self.assertTrue(
            torch.isclose(
                torch.cat([det.class_ids, det.class_ids]), det_new.class_ids
            ).all()
        )

        self.assertTrue(
            torch.isclose(
                torch.cat([det.track_ids, det.track_ids]), det_new.track_ids
            ).all()
        )


class TestBoxes3D(unittest.TestCase):
    """Test cases VisT Boxes3D."""

    def test_scalabel(self) -> None:
        """Testcase for conversion to / from scalabel."""
        num_dets = 10
        detections = generate_dets3d(num_dets, track_ids=True)
        idx_to_class = {0: "car"}
        class_to_idx = {"car": 0}
        scalabel_dets = detections.to_scalabel(idx_to_class)

        detections_new = Boxes3D.from_scalabel(scalabel_dets, class_to_idx)

        scalabel_dets[0].box3d = None
        dets_with_none = Boxes3D.from_scalabel(scalabel_dets, class_to_idx)
        self.assertTrue(
            torch.isclose(
                dets_with_none.boxes[0], detections_new.boxes[1]
            ).all()
        )

        for det, det_new in zip(detections, detections_new):  # type: ignore
            self.assertTrue(torch.isclose(det.boxes, det_new.boxes).all())
            self.assertTrue(
                torch.isclose(det.class_ids.long(), det_new.class_ids).all()
            )
            self.assertTrue(
                torch.isclose(det.track_ids.long(), det_new.track_ids).all()
            )

        detections_new.track_ids = None
        dets_without_tracks = detections_new.to_scalabel(idx_to_class)
        self.assertTrue(
            all(
                (str(i) == det.id for i, det in enumerate(dets_without_tracks))
            )
        )
        # test 7 DoF
        detections_7dof = detections.clone()
        detections_7dof.boxes = detections_7dof.boxes[
            :, [0, 1, 2, 3, 4, 5, 7, 9]
        ]
        scalabel_dets_7dof = detections_7dof.to_scalabel(  # pylint: disable=no-member,line-too-long
            idx_to_class
        )
        for d in scalabel_dets_7dof:
            assert d.box3d is not None
            self.assertEqual(d.box3d.orientation[0], 0)
            self.assertEqual(d.box3d.orientation[2], 0)

        # 7DoF without score
        detections_7dof.boxes = detections_7dof.boxes[:, :-1]
        scalabel_dets_no_score = detections_7dof.to_scalabel(  # pylint: disable=no-member,line-too-long
            idx_to_class
        )
        for d in scalabel_dets_no_score:
            self.assertIsNone(d.score)
            assert d.box3d is not None
            self.assertEqual(d.box3d.orientation[0], 0)
            self.assertEqual(d.box3d.orientation[2], 0)

        # without score
        detections.boxes = detections.boxes[:, :-1]
        scalabel_dets_no_score = detections.to_scalabel(idx_to_class)
        self.assertTrue(all(d.score is None for d in scalabel_dets_no_score))

    def test_clone(self) -> None:
        """Testcase for cloning a Boxes2D object."""
        num_dets = 10
        detections = generate_dets3d(num_dets, track_ids=True)
        detections_new = detections.clone()

        for det, det_new in zip(detections, detections_new):  # type: ignore
            self.assertTrue(torch.isclose(det.boxes, det_new.boxes).all())
            self.assertTrue(
                torch.isclose(det.class_ids, det_new.class_ids).all()
            )
            self.assertTrue(
                torch.isclose(det.track_ids, det_new.track_ids).all()
            )

    def test_cat(self) -> None:
        """Testcase for concatenating a list of Boxes2D objects."""
        num_dets = 10
        det = generate_dets3d(num_dets, track_ids=True)
        det_new = Boxes3D.cat([det, det])

        self.assertTrue(
            torch.isclose(
                torch.cat([det.boxes, det.boxes]), det_new.boxes
            ).all()
        )

        self.assertTrue(
            torch.isclose(
                torch.cat([det.class_ids, det.class_ids]), det_new.class_ids
            ).all()
        )

        self.assertTrue(
            torch.isclose(
                torch.cat([det.track_ids, det.track_ids]), det_new.track_ids
            ).all()
        )



"""Dataset mapper in vist."""
import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
from detectron2.data.common import MapDataset as D2MapDataset
from pydantic import validator
from pydantic.main import BaseModel
from scalabel.label.typing import Extrinsics as ScalabelExtrinsics
from scalabel.label.typing import Frame, ImageSize
from scalabel.label.typing import Intrinsics as ScalabelIntrinsics
from scalabel.label.typing import Label
from scalabel.label.utils import (
    check_crowd,
    check_ignored,
    get_matrix_from_extrinsics,
    get_matrix_from_intrinsics,
)

from ..common.io import DataBackendConfig, build_data_backend
from ..struct import (
    Boxes2D,
    Boxes3D,
    Extrinsics,
    Images,
    InputSample,
    Intrinsics,
    NDArrayUI8,
)
from .transforms import AugmentationConfig, AugParams, build_augmentations
from .utils import im_decode, transform_bbox

__all__ = ["DatasetMapper", "MapDataset"]
logger = logging.getLogger(__name__)


class ReferenceSamplingConfig(BaseModel):
    """Config for customizing the sampling of reference views."""

    type: str = "uniform"
    num_ref_imgs: int = 0
    scope: int = 1
    frame_order: str = "key_first"
    skip_nomatch_samples: bool = False

    @validator("scope")
    def validate_scope(  # type: ignore # pylint: disable=no-self-argument,no-self-use, line-too-long
        cls, value: int, values
    ) -> int:
        """Check scope attribute."""
        if not value > values["num_ref_imgs"] // 2:
            raise ValueError("Scope must be higher than num_ref_imgs / 2.")
        return value


class MapDataset(D2MapDataset):  # type: ignore
    """Map a function over the elements in a dataset."""

    def __init__(  # type: ignore
        self, sampling_cfg: ReferenceSamplingConfig, training, *args, **kwargs
    ):
        """Init."""
        super().__init__(*args, **kwargs)
        self.video_to_indices: Dict[str, List[int]] = defaultdict(list)
        self.frame_to_indices: Dict[str, Dict[int, int]] = defaultdict(dict)
        self._create_video_mapping()
        self.sampling_cfg = sampling_cfg
        self.training = training

    def _create_video_mapping(self) -> None:
        """Create a mapping that returns all img idx for a given video id."""
        for idx, entry in enumerate(self._dataset):
            if entry["video_name"] is not None:
                self.video_to_indices[entry["video_name"]].append(idx)

    def sample_ref_idcs(self, video: str, key_dataset_index: int) -> List[int]:
        """Sample reference dataset indices given video and keyframe index."""
        dataset_indices = self.video_to_indices[video]
        key_index = dataset_indices.index(key_dataset_index)

        if self.sampling_cfg.type == "uniform":
            left = max(0, key_index - self.sampling_cfg.scope)
            right = min(
                key_index + self.sampling_cfg.scope, len(dataset_indices) - 1
            )
            valid_inds = (
                dataset_indices[left:key_index]
                + dataset_indices[key_index + 1 : right + 1]
            )
            ref_dataset_indices = np.random.choice(
                valid_inds, self.sampling_cfg.num_ref_imgs, replace=False
            ).tolist()  # type: List[int]
        elif self.sampling_cfg.type == "sequential":
            right = key_index + 1 + self.sampling_cfg.num_ref_imgs
            if right <= len(dataset_indices):
                ref_dataset_indices = dataset_indices[key_index + 1 : right]
            else:
                left = key_index - (right - len(dataset_indices))
                ref_dataset_indices = (
                    dataset_indices[left:key_index]
                    + dataset_indices[key_index + 1 :]
                )
        else:
            raise NotImplementedError(
                f"Reference view sampling {self.sampling_cfg.type} not "
                f"implemented."
            )

        return ref_dataset_indices

    def sort_samples(
        self, input_samples: List[InputSample]
    ) -> List[InputSample]:
        """Sort samples according to sampling cfg."""
        if self.sampling_cfg.frame_order == "key_first":
            return input_samples
        if self.sampling_cfg.frame_order == "temporal":
            return sorted(
                input_samples,
                key=lambda x: x.metadata.frame_index
                if x.metadata.frame_index is not None
                else 0,
            )
        raise NotImplementedError(
            f"Frame ordering {self.sampling_cfg.frame_order} not "
            f"implemented."
        )

    @staticmethod
    def has_matches(
        key_data: InputSample, ref_data: List[InputSample]
    ) -> bool:
        """Check if key / ref data have matches."""
        has_match = False
        assert key_data.boxes2d is not None
        key_track_ids = key_data.boxes2d.track_ids
        for ref_view in ref_data:
            assert isinstance(ref_view.boxes2d, Boxes2D)
            ref_track_ids = ref_view.boxes2d.track_ids
            match = key_track_ids.view(-1, 1) == ref_track_ids.view(1, -1)
            if match.any():
                has_match = True
                break
        return has_match

    def __getitem__(self, idx: int) -> List[InputSample]:
        """Fully prepare a sample for training/inference."""
        retry_count = 0
        cur_idx = int(idx)

        while True:
            data = self._map_func(self._dataset[cur_idx])
            if data is not None:
                input_data, parameters = data
                if input_data.metadata.attributes is None:
                    input_data.metadata.attributes = dict()
                if self.training:
                    input_data.metadata.attributes["keyframe"] = True
                self._fallback_candidates.add(cur_idx)

                if self.training and self.sampling_cfg.num_ref_imgs > 0:
                    # sample reference views
                    vid_id = input_data.metadata.video_name
                    if vid_id is not None:
                        ref_data = [
                            self._map_func(
                                self._dataset[ref_idx], parameters=parameters
                            )[0]
                            for ref_idx in self.sample_ref_idcs(
                                vid_id, cur_idx
                            )
                        ]
                    else:
                        ref_data = [  # pragma: no cover
                            input_data
                            for _ in range(self.sampling_cfg.num_ref_imgs)
                        ]

                    if (
                        not self.sampling_cfg.skip_nomatch_samples
                        or self.has_matches(input_data, ref_data)
                    ):
                        return self.sort_samples([input_data] + ref_data)
                else:
                    return [input_data]

            # _map_func fails for this idx, use a random new index from the
            # pool
            retry_count += 1
            self._fallback_candidates.discard(cur_idx)
            cur_idx = self._rng.sample(self._fallback_candidates, k=1)[0]

            if retry_count >= 5:
                logger.warning(
                    "Failed to apply `_map_func` for idx: %s, retry count: %s",
                    idx,
                    retry_count,
                )


class DataloaderConfig(BaseModel):
    """Config for dataloader."""

    data_backend: DataBackendConfig = DataBackendConfig()
    workers_per_gpu: int
    fields_to_load: List[str] = ["boxes2d"]
    categories: Optional[List[str]] = None
    skip_empty_samples: bool = False
    clip_bboxes_to_image: bool = True
    compute_global_instance_ids: bool = False
    train_augmentations: Optional[List[AugmentationConfig]] = None
    test_augmentations: Optional[List[AugmentationConfig]] = None
    ref_sampling_cfg: ReferenceSamplingConfig
    image_channel_mode: str
    ignore_unknown_cats: bool = False


class DatasetMapper:
    """DatasetMapper class for VisT.

    A callable which takes a data sample in scalabel format, and maps it into
    a format used by the VisT model. The callable does the following:
    1. Read image from "url"
    2. Applies transforms to the image and annotations
    3. Put data and annotations in VisT format (InputSample)
    """

    def __init__(
        self,
        loader_cfg: DataloaderConfig,
        is_train: bool = True,
    ) -> None:
        """Init."""
        if is_train:
            augs = build_augmentations(loader_cfg.train_augmentations)
        else:
            augs = build_augmentations(loader_cfg.test_augmentations)
        logger.info("Augmentations used: %s", augs)

        self.augs = augs
        self.is_train = is_train
        self.loader_cfg = loader_cfg
        self.data_backend = build_data_backend(loader_cfg.data_backend)
        self.skip_empty_samples = loader_cfg.skip_empty_samples
        self.fields_to_load = loader_cfg.fields_to_load
        for field in self.fields_to_load:
            assert field in ["boxes2d", "boxes3d", "intrinsics", "extrinsics"]

    def load_image(
        self,
        sample: Frame,
    ) -> NDArrayUI8:
        """Load image according to data_backend."""
        assert sample.url is not None
        im_bytes = self.data_backend.get(sample.url)
        image = im_decode(im_bytes, mode=self.loader_cfg.image_channel_mode)
        sample.size = ImageSize(width=image.shape[1], height=image.shape[0])
        return image

    def transform_image(
        self,
        image: NDArrayUI8,
        parameters: Optional[List[AugParams]] = None,
    ) -> Tuple[Images, List[AugParams], torch.Tensor]:
        """Apply image augmentations and convert to torch tensor."""
        image = torch.as_tensor(
            np.ascontiguousarray(image.transpose(2, 0, 1)),
            dtype=torch.float32,
        ).unsqueeze(0)

        if parameters is None:
            parameters = []
        else:
            assert len(parameters) == len(self.augs), (
                "Length of augmentation parameters must equal the number of "
                "augmentations!"
            )

        transform_matrix = torch.eye(3)
        for i, aug in enumerate(self.augs):
            if len(parameters) < len(self.augs):
                parameters.append(aug.forward_parameters(image.shape))
            image, tm = aug(image, parameters[i], return_transform=True)
            transform_matrix = torch.mm(tm[0], transform_matrix)

        image_processed = Images(image, [(image.shape[3], image.shape[2])])
        return image_processed, parameters, transform_matrix

    def transform_annotation(
        self,
        input_sample: InputSample,
        labels: Optional[List[Label]],
        transform_matrix: torch.Tensor,
    ) -> None:
        """Transform annotations."""
        labels_used = []
        if labels is not None:
            cat_dict = dict()
            for label in labels:
                assert label.attributes is not None
                assert label.category is not None
                if not check_crowd(label) and not check_ignored(label):
                    labels_used.append(label)
                    if label.category not in cat_dict:
                        cat_dict[label.category] = int(
                            label.attributes["category_id"]
                        )

        if "boxes2d" in self.fields_to_load:
            if labels is not None and labels_used:
                boxes2d = Boxes2D.from_scalabel(labels_used, cat_dict)
                boxes2d.boxes[:, :4] = transform_bbox(
                    transform_matrix,
                    boxes2d.boxes[:, :4],
                )
                if self.loader_cfg.clip_bboxes_to_image:
                    boxes2d.clip(input_sample.image.image_sizes[0])

                input_sample.boxes2d = boxes2d

        if "boxes3d" in self.fields_to_load:
            if labels is not None and labels_used:
                boxes3d = Boxes3D.from_scalabel(labels_used, cat_dict)
                input_sample.boxes3d = boxes3d

    @staticmethod
    def transform_intrinsics(
        intrinsics: ScalabelIntrinsics, transform_matrix: torch.Tensor
    ) -> Intrinsics:
        """Transform intrinsic camera matrix according to augmentations."""
        intrinsic_matrix = torch.from_numpy(
            get_matrix_from_intrinsics(intrinsics)
        ).to(torch.float32)
        return Intrinsics(torch.mm(transform_matrix, intrinsic_matrix))

    @staticmethod
    def transform_extrinsics(extrinsics: ScalabelExtrinsics) -> Extrinsics:
        """Transform extrinsics from Scalabel to VisT."""
        extrinsics_matrix = torch.from_numpy(
            get_matrix_from_extrinsics(extrinsics)
        ).to(torch.float32)
        return Extrinsics(extrinsics_matrix)

    def __call__(  # type: ignore
        self,
        sample_dict: Dict[str, Any],
        parameters: Optional[List[AugParams]] = None,
    ) -> Optional[Tuple[InputSample, Optional[List[AugParams]]]]:
        """Prepare a single sample in detect format.

        Args:
            sample_dict (serialized Frame): Metadata of one image, in scalabel
            format. Serialized as dict due to multi-processing.
            parameters (List[AugParams]): Augmentation parameter list.

        Returns:
            InputSample: Data format that the model accepts.
            List[AugParams]: augmentation parameters, s.t. ref views can be
            augmented with the same parameters.
        """
        sample = Frame(**sample_dict)

        # image loading, augmentation / to torch.tensor
        image, parameters, transform_matrix = self.transform_image(
            self.load_image(sample),
            parameters=parameters,
        )
        input_data = InputSample(sample, image)

        if (
            sample.intrinsics is not None
            and "intrinsics" in self.fields_to_load
        ):
            input_data.intrinsics = self.transform_intrinsics(
                sample.intrinsics, transform_matrix
            )

        if (
            sample.extrinsics is not None
            and "extrinsics" in self.fields_to_load
        ):
            input_data.extrinsics = self.transform_extrinsics(
                sample.extrinsics
            )

        if not self.is_train:
            del sample.labels
            return input_data, parameters

        self.transform_annotation(input_data, sample.labels, transform_matrix)
        del sample.labels

        if (
            self.skip_empty_samples
            and len(input_data.boxes2d) == 0
            and parameters is None
        ):
            return None  # pragma: no cover

        return input_data, parameters



"""Track graph optimization module."""
from .base import BaseTrackGraph, TrackGraphConfig, build_track_graph
from .quasi_dense import QDTrackGraph

__all__ = [
    "BaseTrackGraph",
    "QDTrackGraph",
    "build_track_graph",
    "TrackGraphConfig",
]



"""Track visualziation test cases."""
import unittest

import numpy as np
import torch

from vist.unittest.utils import generate_dets

from .track import draw_sequence
from .utils import preprocess_image


class TestTrackVis(unittest.TestCase):
    """Tests for vist.vis.track."""

    def test_draw_sequence(self) -> None:
        """Test draw sequence function."""
        seq_imgs = [torch.rand(3, 128, 128) for _ in range(10)]
        proc_imgs = [preprocess_image(im) for im in seq_imgs]
        seq_dets = [
            generate_dets(128, 128, 10, track_ids=True) for _ in range(10)
        ]
        seq = draw_sequence(seq_imgs, seq_dets)

        self.assertEqual(len(seq), 10)

        for orig_frame, frame in zip(proc_imgs, seq):
            self.assertEqual(np.array(frame).shape, (128, 128, 3))
            # check if frame was modified
            self.assertFalse((np.array(orig_frame) == np.array(frame)).all())



"""Evaluation components for tracking."""
import copy
import datetime
import logging
import os
import time
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from typing import Dict, Generator, List, Optional

import torch
from detectron2.data import DatasetCatalog, MetadataCatalog
from detectron2.evaluation import DatasetEvaluator, DatasetEvaluators
from detectron2.utils import comm
from detectron2.utils.logger import log_every_n_seconds
from scalabel.eval.detect import evaluate_det
from scalabel.eval.mot import acc_single_video_mot, evaluate_track
from scalabel.label.io import group_and_sort, save
from scalabel.label.typing import Config, Frame

from vist.struct import (
    EvalResult,
    EvalResults,
    InputSample,
    LabelInstance,
    ModelOutput,
)

from ..common.utils.parallel import gather_predictions


def _detect(
    pred: List[Frame],
    gt: List[Frame],
    cfg: Config,
    ignore_unknown_cats: bool,  # pylint: disable=unused-argument
) -> EvalResult:
    """Wrapper for evaluate_det function."""
    return evaluate_det(gt, pred, cfg)


def _track(
    pred: List[Frame], gt: List[Frame], cfg: Config, ignore_unknown_cats: bool
) -> EvalResult:
    """Wrapper for evaluate_track function."""
    return evaluate_track(
        acc_single_video_mot,
        group_and_sort(gt),
        group_and_sort(pred),
        cfg,
        ignore_unknown_cats=ignore_unknown_cats,
    )


_eval_mapping = dict(detect=_detect, track=_track)


@contextmanager
def inference_context(model: torch.nn.Module) -> Generator[None, None, None]:
    """Context for inference.

    The detect is temporarily changed to eval mode and
    restored to previous mode afterwards.
    """
    training_mode = model.training
    model.eval()
    yield
    model.train(training_mode)


def inference_on_dataset(
    model: torch.nn.Module,
    data_loader: torch.utils.data.DataLoader,
    evaluator: DatasetEvaluator,
) -> EvalResults:
    """Runs model on the data_loader and evaluate the metrics with evaluator.

    Also benchmark the inference speed of `model.__call__` accurately.
    The model will be used in eval mode.

    Args:
        model (callable): a callable which takes an object from
            `data_loader` and returns some outputs.

            If it's an nn.Module, it will be temporarily set to `eval` mode.
            If you wish to evaluate a detect in `training` mode instead, you
            wrap the given detect and override its behavior of `.eval()` and
            `.train()`.
        data_loader: an iterable object with a length.
            The elements it generates will be the inputs to the detect.
        evaluator (DatasetEvaluator): the evaluator to run. Use `None` if
        you only want
            to benchmark, but don't want to do any evaluation.

    Returns:
        EvalResults: The return value of `evaluator.evaluate()`
    """
    num_devices = comm.get_world_size()
    logger = logging.getLogger(__name__)
    logger.info("Start inference on %s images", len(data_loader))

    total = len(data_loader)  # inference data loader must have a fixed length
    if evaluator is None:
        # create a no-op evaluator
        evaluator = DatasetEvaluators([])  # pragma: no cover
    evaluator.reset()

    num_warmup = min(5, total - 1)
    start_time = time.perf_counter()
    total_compute_time = 0.0
    with ExitStack() as stack:
        if isinstance(model, torch.nn.Module):
            stack.enter_context(inference_context(model))
        stack.enter_context(torch.no_grad())

        for idx, inputs in enumerate(data_loader):
            if idx == num_warmup:
                start_time = time.perf_counter()
                total_compute_time = 0

            start_compute_time = time.perf_counter()
            outputs = model(inputs)
            if torch.cuda.is_available():
                torch.cuda.synchronize()  # pragma: no cover
            total_compute_time += time.perf_counter() - start_compute_time
            evaluator.process(inputs, outputs)

            iters_after_start = idx + 1 - num_warmup * int(idx >= num_warmup)
            seconds_per_img = total_compute_time / iters_after_start
            if idx >= num_warmup or seconds_per_img > 5:
                total_seconds_per_img = (
                    time.perf_counter() - start_time
                ) / iters_after_start
                eta = datetime.timedelta(
                    seconds=int(total_seconds_per_img * (total - idx - 1))
                )
                log_every_n_seconds(
                    logging.INFO,
                    "Inference done {}/{}. {:.4f} s / img. ETA={}".format(
                        idx + 1, total, seconds_per_img, str(eta)
                    ),
                    n=5,
                )

    # Measure the time only for this worker (before the synchronization
    # barrier)
    total_time = time.perf_counter() - start_time
    total_time_str = str(datetime.timedelta(seconds=total_time))
    # NOTE this format is parsed by grep
    logger.info(
        "Total inference time: {} ({:.6f} s / img per device, "
        "on {} devices)".format(
            total_time_str, total_time / (total - num_warmup), num_devices
        )
    )
    total_compute_time_str = str(
        datetime.timedelta(seconds=int(total_compute_time))
    )
    logger.info(
        "Total inference pure compute time: {} ({:.6f} s / img per device, "
        "on {} devices)".format(
            total_compute_time_str,
            total_compute_time / (total - num_warmup),
            num_devices,
        )
    )

    results = evaluator.evaluate()
    # An evaluator may return None when not in main process.
    # Replace it by an empty dict instead to make it easier for downstream
    # code to handle
    if results is None:
        results = {}  # pragma: no cover
    return results  # type: ignore


class ScalabelEvaluator(DatasetEvaluator):  # type: ignore
    """Evaluate model using metrics supported in salabel (currently AP / MOTA).

    This class will accumulate information of the inputs/outputs (by
    :meth:`process`), and produce evaluation results in the end (by
    :meth:`evaluate`).
    """

    def __init__(
        self,
        dataset_name: str,
        metrics: List[str],
        distributed: bool = True,
        output_dir: Optional[str] = None,
        ignore_unknown_cats: bool = False,
    ) -> None:
        """Init."""
        self._metrics = list(_eval_mapping.keys())
        self._distributed = distributed
        self._output_dir = output_dir
        self._metadata = MetadataCatalog.get(dataset_name)
        self.gts = DatasetCatalog[dataset_name]()
        self._predictions = defaultdict(list)  # type: Dict[str, List[Frame]]
        self.set_metrics(metrics)
        self.ignore_unknown_cats = ignore_unknown_cats

    def reset(self) -> None:
        """Preparation for a new round of evaluation."""
        self._predictions = defaultdict(list)

    def set_metrics(self, metrics: List[str]) -> None:
        """Set metrics to evaluate."""
        for metric in metrics:
            if metric not in _eval_mapping.keys():  # pragma: no cover
                raise KeyError(f"metric {metric} is not supported")
        self._metrics = metrics

    def process(
        self, inputs: List[List[InputSample]], outputs: ModelOutput
    ) -> None:
        """Process the pair of inputs and outputs."""
        for key, output in outputs.items():
            for inp, out in zip(inputs, output):
                prediction = copy.deepcopy(inp[0].metadata)
                out_cpu = out.to(torch.device("cpu"))
                assert isinstance(out_cpu, LabelInstance)
                prediction.labels = out_cpu.to_scalabel(
                    self._metadata.idx_to_class_mapping
                )
                self._predictions[key].append(prediction)

    def evaluate(self) -> EvalResults:
        """Evaluate the performance after processing all input/output pairs."""
        if self._distributed:
            predictions_dict = gather_predictions(self._predictions)
            if not comm.is_main_process():
                return {}  # pragma: no cover
        else:
            predictions_dict = self._predictions  # pragma: no cover

        results = {}
        for key, predictions in predictions_dict.items():
            if self._output_dir:
                os.makedirs(self._output_dir, exist_ok=True)
                file_path = os.path.join(
                    self._output_dir, f"{key}_predictions.json"
                )
                save(file_path, predictions)

            if key in self._metrics:
                results[key] = _eval_mapping[key](
                    predictions,
                    self.gts,
                    self._metadata.metadata_cfg,
                    self.ignore_unknown_cats,
                )

        return results



"""RoI Pooling module base."""
import abc
from typing import List, Tuple

import torch
from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D


class RoIPoolerConfig(BaseModel, extra="allow"):
    """Base RoI pooler config."""

    type: str = Field(...)
    resolution: Tuple[int, int] = Field(...)


class BaseRoIPooler(metaclass=RegistryHolder):
    """Base class for RoI poolers."""

    @abc.abstractmethod
    def pool(
        self, features: List[torch.Tensor], boxes: List[Boxes2D]
    ) -> torch.Tensor:
        """Pool features in input bounding boxes from given feature maps."""
        raise NotImplementedError


def build_roi_pooler(cfg: RoIPoolerConfig) -> BaseRoIPooler:
    """Build an RoI pooler from config."""
    registry = RegistryHolder.get_registry(BaseRoIPooler)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseRoIPooler)
        return module
    raise NotImplementedError(f"RoIPooler {cfg.type} not found.")



"""RoIHead interface for backend."""

import abc
from typing import Any

import torch
from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder


class SimilarityLearningConfig(BaseModel, extra="allow"):
    """Base config for similarity learning."""

    type: str = Field(...)


class BaseSimilarityHead(torch.nn.Module, metaclass=RegistryHolder):  # type: ignore # pylint: disable=line-too-long
    """Base similarity head class."""

    @abc.abstractmethod
    def forward(self, *args: Any, **kwargs: Any) -> Any:  # type: ignore
        """Forward method.

        Process proposals, output predictions and possibly target
        assignments.
        """
        raise NotImplementedError


def build_similarity_head(cfg: SimilarityLearningConfig) -> BaseSimilarityHead:
    """Build a SimilarityHead from config."""
    registry = RegistryHolder.get_registry(BaseSimilarityHead)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseSimilarityHead)
        return module
    raise NotImplementedError(f"RoIHead {cfg.type} not found.")



"""Loss utility functions."""
from typing import Optional

import torch
from torch.nn._reduction import get_enum


def _reduce(loss: torch.Tensor, reduction: str) -> torch.Tensor:
    """Reduce loss based on pytorch reduction logic."""
    value = get_enum(reduction)
    if value == 0:
        return loss
    if value == 1:
        return loss.mean()
    if value == 2:
        return loss.sum()
    raise ValueError(f"Invalid reduction: {reduction}.")


def weight_reduce_loss(
    loss: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    reduction: str = "mean",
    avg_factor: Optional[float] = None,
) -> torch.Tensor:
    """Apply element-wise weight and reduce loss."""
    # if weight is specified, apply element-wise weight
    if weight is not None:
        loss = loss * weight

    # if avg_factor is not specified, just reduce the loss
    if avg_factor is None:
        loss = _reduce(loss, reduction)
    else:
        # if reduction is mean, then average the loss by avg_factor
        if reduction == "mean":
            loss = loss.sum() / avg_factor
        # if reduction is 'none', then do nothing, otherwise raise an error
        elif reduction != "none":  # pragma: no cover
            raise ValueError('avg_factor can not be used with reduction="sum"')
    return loss



"""Interface for using and extending kornia augmentations in VisT.

Kornia augmentations require the following functions:
- generate_parameters: Generate the (random) parameters for computing the
current transformation.
- compute_transformation: Compute a geometric transformation that transforms
the input. Optionally in Kornia but required here, since we need the transform
to maintain valid projective geometry in 3D tracking.
- apply_transform: Apply the transformation to the input.

Reference: https://kornia.readthedocs.io/en/latest/augmentation.base.html
"""
from typing import Dict, List, Optional, Tuple, Union

import torch
from kornia import augmentation as kornia_augmentation
from kornia.augmentation.base import GeometricAugmentationBase2D
from pydantic.main import BaseModel

from vist.common.registry import RegistryHolder

AugParams = Dict[str, torch.Tensor]


class AugmentationConfig(BaseModel):
    """Data augmentation instance config."""

    type: str
    kwargs: Dict[
        str,
        Union[
            bool,
            float,
            str,
            Tuple[float, float],
            Tuple[int, int],
            List[Tuple[int, int]],
        ],
    ]


class BaseAugmentation(GeometricAugmentationBase2D, metaclass=RegistryHolder):  # type: ignore # pylint: disable=line-too-long
    """Subclass kornia Augmentation to support registry."""

    def __repr__(self) -> str:
        """Print class & params, s.t. user can inspect easily via cmd line."""
        return self.__class__.__name__ + f"({super().__repr__()})"

    def generate_parameters(self, batch_shape: torch.Size) -> AugParams:
        """Generate current parameters."""
        raise NotImplementedError

    def compute_transformation(  # pylint: disable=arguments-renamed
        self, inputs: torch.Tensor, params: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Get the corresponding deterministic transform for a given input.

        Args:
            inputs: Image tensor.
            params: parameters needed for computing transformation matrix.

        Returns:
            Transform: Returns the deterministic transform.
        """
        raise NotImplementedError

    def apply_transform(  # pylint: disable=arguments-renamed
        self, inputs: torch.Tensor, params: AugParams, transform: torch.Tensor
    ) -> torch.Tensor:
        """Apply the transformation given parameters and transform."""
        raise NotImplementedError


def build_augmentation(
    cfg: AugmentationConfig,
) -> BaseAugmentation:
    """Build a single augmentation."""
    registry = RegistryHolder.get_registry(GeometricAugmentationBase2D)

    if cfg.type in registry:
        augmentation = registry[cfg.type]
    elif hasattr(kornia_augmentation, cfg.type):
        augmentation = getattr(kornia_augmentation, cfg.type)
    else:
        raise ValueError(f"Augmentation {cfg.type} not known!")
    return augmentation(**cfg.kwargs)  # type: ignore


def build_augmentations(
    cfgs: Optional[List[AugmentationConfig]],
) -> List[BaseAugmentation]:
    """Build a list of augmentations and return these as List."""
    augmentations = []
    if cfgs is not None:
        for aug_cfg in cfgs:
            augmentations.append(build_augmentation(aug_cfg))
    return augmentations



"""Testcases for d2 utils."""
import unittest

from vist.unittest.utils import get_test_file

from .d2_utils import D2TwoStageDetectorConfig, model_to_detectron2


class TestD2Utils(unittest.TestCase):
    """Testcases for d2 utils."""

    def test_model_tod2(self) -> None:
        """Testcase for config to d2 config."""
        cfg = D2TwoStageDetectorConfig(
            type="D2GeneralizedRCNN",
            model_base="faster-rcnn/r50-fpn",
            num_classes=10,
        )
        model_to_detectron2(cfg)
        cfg.model_base = "COCO-Detection/faster_rcnn_R_50_FPN_3x.yaml"
        cfg.override_mapping = True
        det2cfg = model_to_detectron2(cfg)
        self.assertEqual(det2cfg.MODEL.META_ARCHITECTURE, "GeneralizedRCNN")
        cfg.override_mapping = False

        cfg.model_base = get_test_file("test-cfg.yaml")
        det2cfg = model_to_detectron2(cfg)
        self.assertEqual(det2cfg.MODEL.META_ARCHITECTURE, "GeneralizedRCNN")



"""Multi-positive cross entropy loss."""
from typing import Optional

import torch

from .base import BaseLoss, LossConfig
from .utils import weight_reduce_loss


class MultiPosCrossEntropyLoss(BaseLoss):
    """Multi-positive cross entropy loss class."""

    def __init__(self, cfg: LossConfig):
        """Init."""
        super().__init__()
        self.cfg = cfg

    def forward(  # type: ignore # pylint: disable=arguments-differ
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
        weight: Optional[torch.Tensor] = None,
        reduction_override: Optional[str] = None,
        avg_factor: Optional[float] = None,
    ) -> torch.Tensor:
        """Multi-positive cross entropy loss forward."""
        assert pred.size() == target.size()
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = (
            reduction_override
            if reduction_override is not None
            else self.cfg.reduction
        )
        loss_cls = self.cfg.loss_weight * multi_pos_cross_entropy(
            pred,
            target,
            weight,
            reduction=reduction,
            avg_factor=avg_factor,
        )
        return loss_cls


def multi_pos_cross_entropy(
    pred: torch.Tensor,
    target: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    reduction: str = "mean",
    avg_factor: Optional[float] = None,
) -> torch.Tensor:
    """Calculate multi-positive cross-entropy loss."""
    pos_inds = target == 1
    neg_inds = target == 0
    pred_pos = pred * pos_inds.float()
    pred_neg = pred * neg_inds.float()
    # use -inf to mask out unwanted elements.
    pred_pos[neg_inds] = pred_pos[neg_inds] + float("inf")
    pred_neg[pos_inds] = pred_neg[pos_inds] + float("-inf")

    _pos_expand = torch.repeat_interleave(pred_pos, pred.shape[1], dim=1)
    _neg_expand = pred_neg.repeat(1, pred.shape[1])

    x = torch.nn.functional.pad(
        (_neg_expand - _pos_expand), (0, 1), "constant", 0
    )
    loss = torch.logsumexp(x, dim=1)

    # apply weights and do the reduction
    if weight is not None:
        weight = weight.float()
    loss = weight_reduce_loss(
        loss, weight=weight, reduction=reduction, avg_factor=avg_factor
    )

    return loss



"""Random Sampler."""
from typing import List, Tuple

import torch
from detectron2.modeling.sampling import subsample_labels

from vist.struct import Boxes2D

from ..matchers.base import MatchResult
from .base import BaseSampler, SamplerConfig
from .utils import prepare_target


class RandomSampler(BaseSampler):
    """Random sampler class."""

    def __init__(self, cfg: SamplerConfig):
        """Init."""
        super().__init__()
        self.cfg = cfg
        self.bg_label = 0

    def sample(
        self,
        matching: List[MatchResult],
        boxes: List[Boxes2D],
        targets: List[Boxes2D],
    ) -> Tuple[List[Boxes2D], List[Boxes2D]]:
        """Sample boxes randomly."""
        sampled_boxes, sampled_targets = [], []
        for match, box, target in zip(matching, boxes, targets):
            pos_idx, neg_idx = subsample_labels(
                match.assigned_labels,
                self.cfg.batch_size_per_image,
                self.cfg.positive_fraction,
                self.bg_label,
            )
            sampled_idcs = torch.cat([pos_idx, neg_idx], dim=0)

            sampled_boxes.append(box[sampled_idcs])

            sampled_targets.append(
                prepare_target(len(pos_idx), sampled_idcs, target, match)
            )
        return sampled_boxes, sampled_targets



"""VisT model module."""
from .base import BaseModel, BaseModelConfig, build_model
from .quasi_dense_rcnn import QDGeneralizedRCNN

__all__ = [
    "BaseModelConfig",
    "build_model",
    "BaseModel",
    "QDGeneralizedRCNN",
]



"""Detectron2 detector wrapper."""
from typing import Dict, List, Optional, Tuple

import torch
from detectron2.checkpoint import DetectionCheckpointer
from detectron2.modeling import GeneralizedRCNN
from torch.nn.modules.batchnorm import _BatchNorm

from vist.model.detect.d2_utils import (
    D2TwoStageDetectorConfig,
    box2d_to_proposal,
    detections_to_box2d,
    images_to_imagelist,
    model_to_detectron2,
    proposal_to_box2d,
    target_to_instance,
)
from vist.struct import Boxes2D, Images, InputSample, LossesType, ModelOutput

from ..base import BaseModelConfig
from .base import BaseTwoStageDetector


class D2TwoStageDetector(BaseTwoStageDetector):
    """Detectron2 two-stage detector wrapper."""

    def __init__(self, cfg: BaseModelConfig):
        """Init."""
        super().__init__()
        self.cfg = D2TwoStageDetectorConfig(**cfg.dict())
        self.d2_cfg = model_to_detectron2(self.cfg)
        # pylint: disable=too-many-function-args,missing-kwoa
        self.d2_detector = GeneralizedRCNN(self.d2_cfg)
        self.checkpointer = DetectionCheckpointer(self.d2_detector)
        if self.d2_cfg.MODEL.WEIGHTS != "":
            self.checkpointer.load(self.d2_cfg.MODEL.WEIGHTS)
        if self.cfg.set_batchnorm_eval:
            self.set_batchnorm_eval()

    def set_batchnorm_eval(self) -> None:
        """Set all batchnorm layers in backbone to eval mode."""
        for m in self.d2_detector.modules():
            if isinstance(m, _BatchNorm):
                m.eval()

    @property
    def device(self) -> torch.device:
        """Get device where detect input should be moved to."""
        return self.d2_detector.pixel_mean.device

    def preprocess_image(self, batched_inputs: List[InputSample]) -> Images:
        """Batch, pad (standard stride=32) and normalize the input images."""
        images = Images.cat([inp.image for inp in batched_inputs], self.device)
        images.tensor = (
            images.tensor - self.d2_detector.pixel_mean
        ) / self.d2_detector.pixel_std
        return images

    def forward_train(
        self, batch_inputs: List[List[InputSample]]
    ) -> LossesType:
        """Forward pass during training stage.

        Returns a dict of loss tensors.
        """
        inputs = [inp[0] for inp in batch_inputs]

        targets = []
        for x in inputs:
            assert x.boxes2d is not None
            targets.append(x.boxes2d.to(self.device))

        images = self.preprocess_image(inputs)
        features = self.extract_features(images)
        proposals, rpn_losses = self.generate_proposals(
            images, features, targets
        )
        _, detect_losses = self.generate_detections(
            images, features, proposals, targets, compute_detections=False
        )
        return {**rpn_losses, **detect_losses}

    def forward_test(
        self, batch_inputs: List[InputSample], postprocess: bool = True
    ) -> ModelOutput:
        """Forward pass during testing stage.

        Returns predictions for each input.
        """
        images = self.preprocess_image(batch_inputs)
        features = self.extract_features(images)
        proposals, _ = self.generate_proposals(images, features)
        detections, _ = self.generate_detections(images, features, proposals)
        assert detections is not None

        if postprocess:
            for inp, det in zip(batch_inputs, detections):
                ori_wh = (
                    batch_inputs[0].metadata.size.width,  # type: ignore
                    batch_inputs[0].metadata.size.height,  # type: ignore
                )
                self.postprocess(ori_wh, inp.image.image_sizes[0], det)

        return dict(detect=detections)  # type: ignore

    def extract_features(self, images: Images) -> Dict[str, torch.Tensor]:
        """Detector feature extraction stage.

        Return preprocessed images, backbone output features.
        """
        return self.d2_detector.backbone(images.tensor)  # type: ignore

    def generate_proposals(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        targets: Optional[List[Boxes2D]] = None,
    ) -> Tuple[List[Boxes2D], LossesType]:
        """Detector RPN stage.

        Return proposals per image and losses (empty if no targets).
        """
        images_d2 = images_to_imagelist(images)
        is_training = self.d2_detector.proposal_generator.training
        if targets is not None:
            targets = target_to_instance(targets, images.image_sizes)
        else:
            self.d2_detector.proposal_generator.training = False

        proposals, rpn_losses = self.d2_detector.proposal_generator(
            images_d2, features, targets
        )
        self.d2_detector.proposal_generator.training = is_training
        return proposal_to_box2d(proposals), rpn_losses

    def generate_detections(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        proposals: List[Boxes2D],
        targets: Optional[List[Boxes2D]] = None,
        compute_detections: bool = True,
    ) -> Tuple[Optional[List[Boxes2D]], LossesType]:
        """Detector second stage (RoI Head).

        Return losses (empty if no targets) and optionally detections.
        """
        images_d2 = images_to_imagelist(images)
        proposals = box2d_to_proposal(proposals, images.image_sizes)
        is_training = self.d2_detector.roi_heads.training
        if targets is not None:
            targets = target_to_instance(targets, images.image_sizes)
        else:
            self.d2_detector.roi_heads.training = False

        detections, detect_losses = self.d2_detector.roi_heads(
            images_d2,
            features,
            proposals,
            targets,
        )
        self.d2_detector.roi_heads.training = is_training
        if not self.d2_detector.training:
            detections = detections_to_box2d(detections)
        elif compute_detections:
            detections = proposal_to_box2d(detections)  # pragma: no cover
        else:
            detections = None
        return detections, detect_losses



"""RoI Pooling module from detectron2."""
from typing import List

import torch
from detectron2.modeling.poolers import ROIPooler as D2ROIPooler
from detectron2.structures import Boxes

from vist.struct import Boxes2D

from .base import BaseRoIPooler, RoIPoolerConfig


class D2RoIPoolerConfig(RoIPoolerConfig):
    """D2RoIPooler config."""

    pooling_op: str
    strides: List[int]
    sampling_ratio: int


class D2RoIPooler(BaseRoIPooler):
    """detectron2 roi pooling class."""

    cfg_type = D2RoIPoolerConfig

    def __init__(self, cfg: RoIPoolerConfig):
        """Init."""
        super().__init__()
        self.cfg = D2RoIPoolerConfig(**cfg.dict())

        self.roi_pooler = D2ROIPooler(
            output_size=self.cfg.resolution,
            scales=[1 / s for s in self.cfg.strides],
            sampling_ratio=self.cfg.sampling_ratio,
            pooler_type=self.cfg.pooling_op,
        )

    def pool(
        self, features: List[torch.Tensor], boxes: List[Boxes2D]
    ) -> torch.Tensor:
        """detectron2 based roi pooling operation.

        Args:
            features: list of image feature tensors (e.g.
            fpn levels) - NCHW format
            boxes: list of proposals (per image)

        Returns:
            torch.Tensor: NCHW format, where N = num boxes (total),
            HW is roi size, C is feature dim. Boxes are concatenated along
            dimension 0 for all batch elements.
        """
        x = self.roi_pooler(features, [Boxes(b.boxes[:, :4]) for b in boxes])

        return x



"""Utilities for unit tests."""
import inspect
import math
import os
from typing import List

import torch
from detectron2.data import DatasetCatalog, MetadataCatalog

from vist.struct import Boxes2D, Boxes3D


def get_test_file(file_name: str) -> str:
    """Test test file path."""
    return os.path.join(
        os.path.dirname(os.path.abspath(inspect.stack()[1][1])),
        "testcases",
        file_name,
    )


def generate_dets(
    height: int, width: int, num_dets: int, track_ids: bool = False
) -> Boxes2D:
    """Create random detections."""
    state = torch.random.get_rng_state()
    torch.random.set_rng_state(torch.manual_seed(0).get_state())
    rand_max = torch.repeat_interleave(
        torch.tensor([[width, height, width, height, 1.0]]), num_dets, dim=0
    )
    box_tensor = torch.rand(num_dets, 5) * rand_max
    sorted_xy = [
        box_tensor[:, [0, 2]].sort(dim=-1)[0],
        box_tensor[:, [1, 3]].sort(dim=-1)[0],
    ]
    box_tensor[:, :4] = torch.cat(
        [
            sorted_xy[0][:, 0:1],
            sorted_xy[1][:, 0:1],
            sorted_xy[0][:, 1:2],
            sorted_xy[1][:, 1:2],
        ],
        dim=-1,
    )
    tracks = torch.arange(0, num_dets) if track_ids else None
    dets = Boxes2D(box_tensor, torch.zeros(num_dets), tracks)
    torch.random.set_rng_state(state)
    return dets


def generate_dets3d(num_dets: int, track_ids: bool = False) -> Boxes3D:
    """Create random 3D detections."""
    state = torch.random.get_rng_state()
    torch.random.set_rng_state(torch.manual_seed(0).get_state())
    rand_min = torch.repeat_interleave(
        torch.tensor([[-10, -1, 0, 1.5, 1, 3, 0, -math.pi, 0, 1.0]]),
        num_dets,
        dim=0,
    )
    rand_max = torch.repeat_interleave(
        torch.tensor([[10, 1, 80, 2, 2, 4, 0, math.pi, 0, 1.0]]),
        num_dets,
        dim=0,
    )
    box_tensor = torch.rand(num_dets, 10) * (rand_max - rand_min) + rand_min

    tracks = torch.arange(0, num_dets) if track_ids else None
    dets = Boxes3D(box_tensor, torch.zeros(num_dets), tracks)
    torch.random.set_rng_state(state)
    return dets


def d2_data_reset(datasets: List[str]) -> None:
    """Delete all given dataset instances."""
    for ds in datasets:
        if ds in DatasetCatalog.keys():
            DatasetCatalog.remove(ds)
            MetadataCatalog.remove(ds)



"""Config preparation functions for the models module."""
import logging
import os
from typing import List

import torch
from detectron2.config import CfgNode, get_cfg
from detectron2.data import DatasetCatalog
from detectron2.utils import comm
from detectron2.utils.collect_env import collect_env_info
from detectron2.utils.env import seed_all_rng
from detectron2.utils.file_io import PathManager
from detectron2.utils.logger import setup_logger
from devtools import debug

from vist.config import Config, Launch
from vist.data.datasets import register_dataset
from vist.data.datasets.base import BaseDatasetConfig


def _register(datasets: List[BaseDatasetConfig]) -> List[str]:
    """Register datasets in detectron2."""
    names = []
    for dataset in datasets:
        names.append(dataset.name)
        try:
            DatasetCatalog.get(dataset.name)
        except KeyError:
            register_dataset(dataset)
            continue
        logger = logging.getLogger(__name__)
        logger.info(
            "WARNING: You tried to register the same dataset name "
            "twice. Skipping instance:\n%s",
            dataset,
        )
    return names


def register_directory(input_path: str) -> str:
    """Register directory containing input data as dataset."""
    if input_path[-1] == "/":
        input_path = input_path[:-1]
    dataset_name = os.path.basename(input_path)
    dataset = BaseDatasetConfig(
        type="Custom", name=dataset_name, data_root=input_path
    )
    register_dataset(dataset)
    return dataset_name


def to_detectron2(config: Config) -> CfgNode:
    """Convert a Config object to a detectron2 readable configuration."""
    cfg = get_cfg()
    cfg.OUTPUT_DIR = config.launch.output_dir

    # convert solver attributes
    if config.solver is not None:
        cfg.SOLVER.IMS_PER_BATCH = config.solver.images_per_gpu
        if config.launch.num_gpus > 0:
            cfg.SOLVER.IMS_PER_BATCH *= config.launch.num_gpus
        cfg.SOLVER.LR_SCHEDULER_NAME = config.solver.lr_policy
        cfg.SOLVER.BASE_LR = config.solver.base_lr
        cfg.SOLVER.MAX_ITER = config.solver.max_iters
        if config.solver.warmup_iters is not None:
            cfg.SOLVER.WARMUP_ITERS = config.solver.warmup_iters
        if config.solver.steps is not None:
            cfg.SOLVER.STEPS = config.solver.steps
        if config.solver.checkpoint_period is not None:
            cfg.SOLVER.CHECKPOINT_PERIOD = config.solver.checkpoint_period
        if config.solver.eval_period is not None:
            cfg.TEST.EVAL_PERIOD = config.solver.eval_period

    # convert dataloader attributes
    if config.dataloader is not None:
        cfg.DATALOADER.NUM_WORKERS = config.dataloader.workers_per_gpu

    # register datasets
    if config.train is not None:
        cfg.DATASETS.TRAIN = _register(config.train)
    if config.test is not None:
        cfg.DATASETS.TEST = _register(config.test)

    return cfg


def default_setup(cfg: Config, det2cfg: CfgNode, args: Launch) -> None:
    """Perform some basic common setups at the beginning of a job.

    1. Set up the logger
    2. Log basic information about environment, cmdline arguments, and config
    3. Backup the config to the output directory
    """
    rank = comm.get_rank()
    logger = setup_logger(
        det2cfg.OUTPUT_DIR, distributed_rank=rank, name="vist"
    )
    setup_logger(
        os.path.join(det2cfg.OUTPUT_DIR, "d2_log.txt"),
        distributed_rank=rank,
        name="detectron2",
    )

    logger.info(
        "Rank of current process: %s. World size: %s",
        rank,
        comm.get_world_size(),
    )
    logger.info("Environment info: %s", collect_env_info())

    logger.info(
        "Running with full config:\n %s",
        str(debug.format(cfg)).split("\n", 1)[1],
    )
    if comm.is_main_process():
        # Note: some of the detectron2 scripts may expect the existence of
        # config.yaml in output directory
        path = os.path.join(det2cfg.OUTPUT_DIR, "config.yaml")
        with PathManager.open(path, "w") as f:
            f.write(det2cfg.dump())
        # save VisT config
        path = os.path.join(det2cfg.OUTPUT_DIR, "config.json")
        with PathManager.open(path, "w") as f:
            f.write(cfg.json())
        logger.info("Full config saved to %s", path)

    # make sure each worker has different, yet deterministic seed if specified
    if cfg.launch.seed < 0:
        seed_all_rng()
    else:
        logger.info("Using a fixed random seed: %s", cfg.launch.seed)
        seed_all_rng(cfg.launch.seed + rank)

    # cudnn benchmark has large overhead. It shouldn't be used considering the
    # small size of typical validation set.
    if args.action == "train":
        torch.backends.cudnn.benchmark = det2cfg.CUDNN_BENCHMARK



"""Detection utils."""
import os
from typing import Dict, List, Optional, Tuple, Union

import torch
from detectron2 import model_zoo
from detectron2.config import CfgNode, get_cfg
from detectron2.structures import Boxes, ImageList, Instances

from vist.struct import Boxes2D, Images

from ..base import BaseModelConfig

model_mapping = {
    "faster-rcnn": "COCO-Detection/faster_rcnn_",
    "mask-rcnn": "COCO-InstanceSegmentation/mask_rcnn_",
}

backbone_mapping = {
    "r101-fpn": "R_101_FPN_3x.yaml",
    "r101-c4": "R_101_C4_3x.yaml",
    "r101-dc5": "R_101_DC5_3x.yaml",
    "r50-fpn": "R_50_FPN_3x.yaml",
    "r50-c4": "R_50_C4_3x.yaml",
    "r50-dc5": "R_50_DC5_3x.yaml",
}


class D2TwoStageDetectorConfig(BaseModelConfig):
    """Config for detectron2 two stage models."""

    model_base: str
    model_kwargs: Optional[Dict[str, Union[bool, float, str, List[float]]]]
    override_mapping: Optional[bool] = False
    num_classes: Optional[int]
    set_batchnorm_eval: bool = False
    weights: Optional[str]


def detections_to_box2d(detections: List[Instances]) -> List[Boxes2D]:
    """Convert d2 Instances representing detections to Boxes2D."""
    result = []
    for detection in detections:
        boxes, scores, cls = (
            detection.pred_boxes.tensor,
            detection.scores,
            detection.pred_classes,
        )
        result.append(
            Boxes2D(
                torch.cat([boxes, scores.unsqueeze(-1)], -1),
                class_ids=cls,
            )
        )
    return result


def proposal_to_box2d(proposals: List[Instances]) -> List[Boxes2D]:
    """Convert d2 Instances representing proposals to Boxes2D."""
    result = []
    for proposal in proposals:
        boxes, logits = (
            proposal.proposal_boxes.tensor,
            proposal.objectness_logits,
        )
        result.append(
            Boxes2D(
                torch.cat([boxes, logits.unsqueeze(-1)], -1),
            )
        )
    return result


def box2d_to_proposal(
    proposals: List[Boxes2D], imgs_wh: List[Tuple[int, int]]
) -> List[Instances]:
    """Convert Boxes2D representing proposals to d2 Instances."""
    result = []
    for proposal, img_wh in zip(proposals, imgs_wh):
        boxes, logits = (
            proposal.boxes[:, :4],
            proposal.boxes[:, -1],
        )
        fields = dict(proposal_boxes=Boxes(boxes), objectness_logits=logits)
        result.append(Instances((img_wh[1], img_wh[0]), **fields))
    return result


def target_to_instance(
    targets: List[Boxes2D], imgs_wh: List[Tuple[int, int]]
) -> List[Instances]:
    """Convert Boxes2D representing targets to d2 Instances."""
    result = []
    for target, img_wh in zip(targets, imgs_wh):
        boxes, cls, track_ids = (
            target.boxes,
            target.class_ids,
            target.track_ids,
        )
        fields = dict(gt_boxes=Boxes(boxes), gt_classes=cls)
        if track_ids is not None:
            fields["track_ids"] = track_ids
        result.append(Instances((img_wh[1], img_wh[0]), **fields))
    return result


def images_to_imagelist(images: Images) -> ImageList:
    """Convert Images to ImageList (switch from wh to hw for image sizes)."""
    return ImageList(
        images.tensor,
        image_sizes=[(wh[1], wh[0]) for wh in images.image_sizes],
    )


def model_to_detectron2(config: D2TwoStageDetectorConfig) -> CfgNode:
    """Convert a Detector config to a detectron2 readable config."""
    cfg = get_cfg()

    # load detect base config, checkpoint
    d2_model_string = None
    if os.path.exists(config.model_base):
        base_cfg = config.model_base
    else:
        if config.override_mapping:
            d2_model_string = config.model_base
        else:
            model, backbone = config.model_base.split("/")
            d2_model_string = model_mapping[model] + backbone_mapping[backbone]
        base_cfg = model_zoo.get_config_file(d2_model_string)

    cfg.merge_from_file(base_cfg)

    # prepare checkpoint path
    if config.weights is not None:
        if config.weights == "detectron2" and d2_model_string is not None:
            cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url(d2_model_string)
        else:
            cfg.MODEL.WEIGHTS = config.weights
    else:
        cfg.MODEL.WEIGHTS = ""

    # convert detect attributes
    if config.num_classes:
        cfg.MODEL.ROI_HEADS.NUM_CLASSES = config.num_classes
        cfg.MODEL.RETINANET.NUM_CLASSES = config.num_classes

    # add keyword args in config
    if config.model_kwargs:
        for k, v in config.model_kwargs.items():
            attr = cfg
            partial_keys = k.split(".")
            partial_keys, last_key = partial_keys[:-1], partial_keys[-1]
            for part_k in partial_keys:
                attr = attr.get(part_k)
            attr_type = type(attr.get(last_key))
            attr.__setattr__(last_key, attr_type(v))
    return cfg



"""Test cases for random sampler."""
import unittest

import torch

from vist.struct import Boxes2D

from ..matchers.base import MatchResult
from .base import SamplerConfig
from .random import RandomSampler


class TestRandom(unittest.TestCase):
    """Test cases for random sampler."""

    def test_sample(self) -> None:
        """Testcase for sample function."""
        samples_per_img = 10
        pos_fract = 0.5
        num_samples = 10
        num_gts = 3

        sampler = RandomSampler(
            SamplerConfig(
                type="random",
                batch_size_per_image=samples_per_img,
                positive_fraction=pos_fract,
            )
        )
        matching = [
            MatchResult(
                assigned_gt_indices=torch.zeros(num_samples),
                assigned_gt_iou=torch.ones(num_samples),
                assigned_labels=torch.ones(num_samples),
            )
        ]
        boxes = [Boxes2D(torch.rand(num_samples, 5))]
        targets = [Boxes2D(torch.rand(num_gts, 5), torch.zeros(num_gts))]
        sampled_boxes, sampled_targets = sampler.sample(
            matching, boxes, targets
        )
        self.assertEqual(
            len(sampled_boxes[0]), int(samples_per_img * pos_fract)
        )
        self.assertEqual(len(sampled_boxes[0]), len(sampled_targets[0]))

        for target in sampled_targets[0]:  # type: ignore
            self.assertTrue(
                torch.isclose(targets[0][0].boxes, target.boxes).all()
            )

        boxes = [Boxes2D(torch.rand(num_samples, 5))]
        targets = [Boxes2D(torch.empty(0, 5), torch.empty(0), torch.empty(0))]
        sampled_boxes, sampled_targets = sampler.sample(
            matching, boxes, targets
        )
        self.assertEqual(len(sampled_boxes[0]), len(sampled_targets[0]))



"""Init sampler module."""
from .base import BaseSampler, SamplerConfig, build_sampler
from .combined import CombinedSampler
from .random import RandomSampler

__all__ = [
    "BaseSampler",
    "RandomSampler",
    "CombinedSampler",
    "build_sampler",
    "SamplerConfig",
]



"""Hdf5 data backend."""
import os
from typing import Dict

from .base import BaseDataBackend, DataBackendConfig


class HDF5Backend(BaseDataBackend):
    """Backend for loading data from HDF5 files.

    This backend works with filepaths pointing to valid HDF5 files. We assume
    that the given HDF5 file contains the whole dataset associated to this
    backend.

    You can use the provided script at tools/datasets_to_hdf5.py to convert
    your dataset to the expected hdf5 format before using this backend.
    """

    def __init__(self, cfg: DataBackendConfig):
        """Init."""
        super().__init__()
        try:
            import h5py  # pylint: disable=import-outside-toplevel
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "Please install h5py to enable HDF5Backend."
            ) from e
        self.cfg = DataBackendConfig(**cfg.dict())
        self.h5_file_api = h5py.File
        self.is_hdf5 = h5py.is_hdf5
        self.db_cache: Dict[str, h5py.File] = dict()

    def get(self, filepath: str) -> bytes:
        """Get values according to the filepath as bytes."""
        filepath_as_list = filepath.split("/")
        keys = []

        while filepath and not self.is_hdf5(filepath):
            keys.append(filepath_as_list.pop())
            filepath = "/".join(filepath_as_list)

        if not os.path.exists(filepath):
            raise FileNotFoundError(
                f"Corresponding HDF5 file not found:" f" {filepath}"
            )

        if filepath not in self.db_cache.keys():
            client = self.h5_file_api(filepath, "r")
            self.db_cache[filepath] = client
        else:
            client = self.db_cache[filepath]

        url = "/".join(reversed(keys))
        value_buf = client
        while keys:
            value_buf = value_buf.get(keys.pop())
            if value_buf is None:
                raise ValueError(f"Value {url} not found in {filepath}!")

        return bytes(value_buf[()])



"""Utils for samplers."""
from typing import Tuple

import torch

from vist.common.bbox.matchers import MatchResult
from vist.struct import Boxes2D


def prepare_target(
    num_pos: int,
    sampled_idcs: torch.Tensor,
    target: Boxes2D,
    match: MatchResult,
) -> Boxes2D:
    """Prepare target from sampled indices."""
    if len(target):
        sampled_target = target[match.assigned_gt_indices.long()[sampled_idcs]]
        assert sampled_target.class_ids is not None, "Targets have no class"
        sampled_target.class_ids[num_pos:] = -1
        if sampled_target.track_ids is not None:
            sampled_target.track_ids[num_pos:] = -1
    else:
        class_ids = torch.ones(len(sampled_idcs)).to(target.device) * -1
        track_ids = None
        if target.track_ids is not None:
            track_ids = class_ids.clone()
        sampled_target = Boxes2D(
            torch.zeros(len(sampled_idcs), 5).to(target.device),
            class_ids,
            track_ids,
        )
    return sampled_target


def nonzero_tuple(
    tensor: torch.Tensor,
) -> Tuple[torch.Tensor]:  # pragma: no cover
    """A 'as_tuple=True' version of torch.nonzero to support torchscript.

    because of https://github.com/pytorch/pytorch/issues/38718
    """
    if torch.jit.is_scripting():
        if tensor.dim() == 0:
            return tensor.unsqueeze(0).nonzero().unbind(1)  # type: ignore
        return tensor.nonzero().unbind(1)  # type: ignore
    return tensor.nonzero(as_tuple=True)  # type: ignore



"""Tracking visualziations."""
from typing import List, Sequence, Union

import torch
from PIL import Image

from vist.struct import Boxes2D

from .image import draw_bbox, imshow_bboxes
from .utils import BoxType, ImageType, preprocess_boxes, preprocess_image


def draw_sequence(
    frames: List[Union[ImageType, Image.Image]],
    boxes: Sequence[BoxType],
    mode: str = "BGR",
) -> List[Image.Image]:
    """Draw predictions of a complete sequence."""
    processed_frames = [
        preprocess_image(f, mode) if not isinstance(f, Image.Image) else f
        for f in frames
    ]
    for frame, boxes2d in zip(processed_frames, boxes):
        box_list, col_list, label_list = preprocess_boxes(boxes2d)
        for box, col, label in zip(box_list, col_list, label_list):
            draw_bbox(frame, box, col, label)
    return processed_frames


def visualize_matches(
    key_inputs: List[ImageType],
    ref_inputs: List[ImageType],
    key_boxes: List[Boxes2D],
    ref_boxes: List[Boxes2D],
    mode: str = "BGR",
) -> None:  # pragma: no cover
    """Visualize paired bounding boxes successively for batched frame pairs."""
    for batch_i, (key_box, ref_box) in enumerate(zip(key_boxes, ref_boxes)):
        target = (
            key_box.track_ids.view(-1, 1) == ref_box.track_ids.view(1, -1)
        ).int()
        for key_i in range(target.shape[0]):
            if target[key_i].sum() > 0:
                ref_i = torch.argmax(target[key_i]).item()
                print("key view")
                imshow_bboxes(key_inputs[batch_i], key_box[key_i], mode)
                print("ref view")
                imshow_bboxes(ref_inputs[batch_i], ref_box[ref_i], mode)



"""VisT data Samplers."""
from typing import Generator

from detectron2.utils import comm
from torch.utils.data.sampler import Sampler

from .dataset_mapper import MapDataset


class TrackingInferenceSampler(Sampler):  # type: ignore
    """Produce sequence ordered indices for inference across all workers.

    Inference needs to run on the __exact__ set of sequences and their
    respecitve samples, therefore if the sequences are not divible by the
    number of workers of if they have different length, the sampler
    produces different number of samples on different workers.
    """

    def __init__(self, dataset: MapDataset):
        """Init."""
        super().__init__(None)
        self._sequences = list(dataset.video_to_indices.keys())
        self._num_seqs = len(self._sequences)
        assert self._num_seqs > 0
        self._rank = comm.get_rank()
        self._world_size = comm.get_world_size()

        shard_size = (self._num_seqs - 1) // self._world_size + 1
        begin = shard_size * self._rank
        end = min(shard_size * (self._rank + 1), self._num_seqs)
        self._local_idcs = []
        for i in range(begin, end):
            self._local_idcs.extend(
                dataset.video_to_indices[self._sequences[i]]
            )

    def __iter__(self) -> Generator[int, None, None]:
        """Iteration method."""
        yield from self._local_idcs

    def __len__(self) -> int:
        """Return length of sampler instance."""
        return len(self._local_idcs)



"""Utilities for timing."""
from time import perf_counter
from typing import no_type_check


@no_type_check
def timeit(func):
    """Function to be used as decorator to time a function."""

    def timed(*args, **kwargs):
        tic = perf_counter()
        result = func(*args, **kwargs)
        toc = perf_counter()
        print("%r  %2.2f ms" % (func.__name__, (toc - tic) * 1000))
        return result

    return timed



"""Package for data transformations / augmentations."""
from .augmentations import Resize
from .base import AugmentationConfig, AugParams, build_augmentations

__all__ = [
    "build_augmentations",
    "AugmentationConfig",
    "AugParams",
    "Resize",
]



"""Utils for parallel processing."""
import itertools
from typing import Dict, List

from detectron2.utils import comm
from scalabel.label.typing import Frame


def gather_predictions(
    predictions: Dict[str, List[Frame]]
) -> Dict[str, List[Frame]]:
    """Gather prediction dict in distributed setting."""
    comm.synchronize()
    predictions_list = comm.gather(predictions, dst=0)

    result = {}
    for key in predictions:
        prediction_list = [p[key] for p in predictions_list]
        result[key] = list(itertools.chain(*prediction_list))
    return result



"""Visualization utils test cases."""
import math
import unittest

import numpy as np
import torch
from PIL import Image

from vist.unittest.utils import generate_dets

from .utils import (
    box3d_to_corners,
    generate_colors,
    preprocess_boxes,
    preprocess_image,
)


class TestUtils(unittest.TestCase):
    """Testcase class vis utils."""

    def test_generate_colors(self) -> None:
        """Test generate colors method."""
        cols = generate_colors(10)
        self.assertEqual(len(cols), 10)
        for col in cols:
            self.assertEqual(len(col), 3)
            for val in col:
                self.assertTrue(isinstance(val, int))
                self.assertTrue(0 <= val < 256)

    def test_preprocess_image(self) -> None:
        """Test preprocess_image method."""
        img = torch.rand((3, 128, 128))
        proc_img = preprocess_image(img)
        self.assertTrue(isinstance(proc_img, Image.Image))
        proc_img = np.array(proc_img)
        self.assertEqual(proc_img.shape, (128, 128, 3))
        self.assertTrue(np.min(proc_img) >= 0)  # type: ignore
        self.assertTrue(np.max(proc_img) < 256)  # type: ignore

    def test_preprocess_boxes(self) -> None:
        """Test preprocess_boxes method."""
        dets = [generate_dets(128, 128, 10, track_ids=True)]
        dets[0].track_ids = dets[0].track_ids.unsqueeze(-1)

        # with score
        proc_dets, cols, labels = preprocess_boxes(dets)
        self.assertTrue(
            len(dets[0]) == len(proc_dets) == len(cols) == len(labels)
        )
        self.assertTrue(len(cols) == len(set(cols)))
        for det, proc_det, label in zip(dets[0], proc_dets, labels):  # type: ignore # pylint: disable=line-too-long
            det_box = det.boxes[0, :4].numpy().tolist()
            self.assertEqual(det_box, proc_det)
            self.assertEqual(0, int(label[2]))

        # without score
        dets[0].boxes = dets[0].boxes[:, :4]
        proc_dets, cols, labels = preprocess_boxes(dets)
        self.assertTrue(
            len(dets[0]) == len(proc_dets) == len(cols) == len(labels)
        )
        self.assertTrue(len(cols) == len(set(cols)))
        for det, proc_det, label in zip(dets[0], proc_dets, labels):  # type: ignore # pylint: disable=line-too-long
            det_box = det.boxes[0, :4].numpy().tolist()
            self.assertEqual(det_box, proc_det)
            self.assertEqual(0, int(label[2]))

        dets[0].track_ids = None
        proc_dets, cols, _ = preprocess_boxes(dets)
        self.assertTrue(len(set(cols)) == 1)

        dets[0].class_ids = None
        proc_dets, cols, _ = preprocess_boxes(dets)
        self.assertTrue(len(set(cols)) == 1)

    def test_box3d_to_corners(self) -> None:
        """Test for box3d_to_corners function."""
        box_corners = [
            [11.0, 11.0, 8.0],
            [9.0, 11.0, 8.0],
            [9.0, 11.0, 12.0],
            [11.0, 11.0, 12.0],
            [11.0, 9.0, 8.0],
            [9.0, 9.0, 8.0],
            [9.0, 9.0, 12.0],
            [11.0, 9.0, 12.0],
        ]
        corners = box3d_to_corners([10, 10, 10, 2, 2, 4, math.pi / 2])
        self.assertTrue((corners == box_corners).all())

        corners = box3d_to_corners([10, 10, 10, 2, 2, 4, 0, math.pi / 2, 0])
        self.assertTrue((corners == box_corners).all())



"""Base class for VisT detectors."""

import abc
from typing import Dict, List, Optional, Tuple, Union

import torch

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D, Images, InputSample, LossesType, ModelOutput

from ..base import BaseModel


class BaseDetector(BaseModel, metaclass=RegistryHolder):
    """Base detector class."""

    @abc.abstractmethod
    def preprocess_image(self, batched_inputs: List[InputSample]) -> Images:
        """Normalize, pad and batch the input images."""
        raise NotImplementedError

    def forward(
        self, batch_inputs: List[List[InputSample]]
    ) -> Union[LossesType, ModelOutput]:
        """Model forward function."""
        if self.training:
            assert all(
                len(inp) == 1 for inp in batch_inputs
            ), "No reference views allowed in detector training!"
            return self.forward_train(batch_inputs)
        inputs = [inp[0] for inp in batch_inputs]  # no ref views during test
        return self.forward_test(inputs)

    @abc.abstractmethod
    def extract_features(self, images: Images) -> Dict[str, torch.Tensor]:
        """Detector feature extraction stage.

        Return backbone output features
        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_detections(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        proposals: List[Boxes2D],
        targets: Optional[List[Boxes2D]] = None,
        compute_detections: bool = True,
    ) -> Tuple[Optional[List[Boxes2D]], LossesType]:
        """Detector second stage (RoI Head).

        Return losses (empty if no targets) and optionally detections.
        """
        raise NotImplementedError


class BaseTwoStageDetector(BaseDetector):
    """Base class for two-stage detectors."""

    @abc.abstractmethod
    def generate_proposals(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        targets: Optional[List[Boxes2D]] = None,
    ) -> Tuple[List[Boxes2D], LossesType]:
        """Detector RPN stage.

        Return proposals per image and losses (empty if no targets).
        """
        raise NotImplementedError



"""DefaultTrainer for VisT."""
import logging
import os
from collections import OrderedDict
from typing import Dict, List, Optional
from typing import OrderedDict as OrderedDictType

import torch
from detectron2.checkpoint import Checkpointer
from detectron2.config import CfgNode
from detectron2.engine import DefaultTrainer as D2DefaultTrainer
from detectron2.engine import PeriodicWriter
from detectron2.evaluation import DatasetEvaluator
from detectron2.utils.comm import is_main_process

from vist.config import Config
from vist.data import build_test_loader, build_train_loader
from vist.model import build_model
from vist.struct import EvalResults
from vist.vis import ScalabelVisualizer

from .evaluator import ScalabelEvaluator, inference_on_dataset
from .utils import default_setup, register_directory, to_detectron2


class DefaultTrainer(D2DefaultTrainer):  # type: ignore
    """VisT DefaultTrainer class."""

    def __init__(self, cfg: Config, det2cfg: CfgNode):
        """Init."""
        self.vist_cfg = cfg
        super().__init__(det2cfg)
        # Update hooks with custom parameters / objects
        logp = cfg.solver.log_period
        for hook in self._hooks:
            if logp is not None and isinstance(hook, PeriodicWriter):
                hook._period = logp  # pylint: disable=protected-access

    def build_model(self, cfg: CfgNode) -> torch.nn.Module:
        """Builds model."""
        model = build_model(self.vist_cfg.model)
        if hasattr(model, "detector") and hasattr(model.detector, "d2_cfg"):
            cfg.MODEL.merge_from_other_cfg(model.detector.d2_cfg.MODEL)
        model.to(torch.device(self.vist_cfg.launch.device))
        logger = logging.getLogger(__name__)
        logger.info("Model:\n%s", model)
        return model

    def build_train_loader(self, cfg: CfgNode) -> torch.utils.data.DataLoader:
        """Calls :func:`vist.data.build_train_loader`."""
        return build_train_loader(self.vist_cfg.dataloader, cfg)

    def build_test_loader(
        self, cfg: CfgNode, dataset_name: str
    ) -> torch.utils.data.DataLoader:
        """Calls static version."""
        return self.build_test_loader_static(
            self.vist_cfg, dataset_name
        )  # pragma: no cover

    @classmethod
    def build_test_loader_static(
        cls, vist_cfg: Config, dataset_name: str
    ) -> torch.utils.data.DataLoader:
        """Calls :func:`vist.data.build_test_loader`."""
        sampling = "sequence_based"
        for ds in vist_cfg.test:
            if ds.name == dataset_name:
                sampling = ds.inference_sampling
        return build_test_loader(vist_cfg.dataloader, dataset_name, sampling)

    def build_evaluator(
        self, cfg: CfgNode, dataset_name: str
    ) -> DatasetEvaluator:
        """Build Scalabel evaluators."""
        return self.build_evaluator_static(
            self.vist_cfg, cfg, dataset_name
        )  # pragma: no cover

    @classmethod
    def build_evaluator_static(
        cls, vist_cfg: Config, cfg: CfgNode, dataset_name: str
    ) -> DatasetEvaluator:
        """Build evaluators."""
        output_folder = os.path.join(cfg.OUTPUT_DIR, dataset_name)
        ignore_unknown_cats = vist_cfg.dataloader.ignore_unknown_cats
        metrics = [
            ds.eval_metrics for ds in vist_cfg.test if ds.name == dataset_name
        ][0]
        evaluator = ScalabelEvaluator(
            dataset_name, metrics, True, output_folder, ignore_unknown_cats
        )
        return evaluator

    def train(self) -> Dict[str, EvalResults]:
        """Run training."""
        super().train()
        return self._last_eval_results  # type: ignore

    def test(
        self,
        cfg: CfgNode,
        model: torch.nn.Module,
        evaluators: Optional[List[DatasetEvaluator]] = None,
    ) -> Dict[str, EvalResults]:
        """Calls static test function."""
        return self.test_static(self.vist_cfg, cfg, model, evaluators)

    @classmethod
    def test_static(
        cls,
        vist_cfg: Config,
        cfg: CfgNode,
        model: torch.nn.Module,
        evaluators: Optional[List[DatasetEvaluator]] = None,
    ) -> OrderedDictType[str, EvalResults]:
        """Test model with given evaluators."""
        logger = logging.getLogger(__name__)
        assert vist_cfg.test is not None
        datasets = [ds.name for ds in vist_cfg.test]
        if isinstance(evaluators, DatasetEvaluator):
            evaluators = [evaluators]  # pragma: no cover
        if evaluators is not None:
            assert len(cfg.DATASETS.TEST) == len(  # pragma: no cover
                evaluators
            ), "{} != {}".format(len(datasets), len(evaluators))

        results = OrderedDict()  # type: ignore
        for idx, dataset_name in enumerate(datasets):
            data_loader = cls.build_test_loader_static(vist_cfg, dataset_name)
            # When evaluators are passed in as arguments, implicitly assume
            # that evaluators can be created before data_loader.
            if evaluators is not None:
                evaluator = evaluators[idx]  # pragma: no cover
            else:
                try:
                    evaluator = cls.build_evaluator_static(
                        vist_cfg, cfg, dataset_name
                    )
                except NotImplementedError:  # pragma: no cover
                    logger.warning(
                        "No evaluator found. Use `Trainer.test(evaluators=)`, "
                        "or implement its `build_evaluator` method."
                    )
                    results[dataset_name] = {}
                    continue
            results_i = inference_on_dataset(model, data_loader, evaluator)
            results[dataset_name] = results_i
            if is_main_process():
                assert isinstance(results_i, dict), (
                    "Evaluator must return a dict on the main process. Got {} "
                    "instead.".format(results_i)
                )

        if len(results) == 1:
            results = list(results.values())[0]  # type: ignore
        return results

    @classmethod
    def predict(
        cls,
        vist_cfg: Config,
        model: torch.nn.Module,
    ) -> None:
        """Test detect with given evaluators."""
        assert vist_cfg.launch.output_dir is not None
        if vist_cfg.launch.input_dir is not None:
            datasets = [register_directory(vist_cfg.launch.input_dir)]
        else:
            assert vist_cfg.test is not None
            datasets = [ds.name for ds in vist_cfg.test]

        for dataset_name in datasets:
            data_loader = cls.build_test_loader_static(vist_cfg, dataset_name)
            output_folder = os.path.join(
                vist_cfg.launch.output_dir, dataset_name
            )
            visualizer = ScalabelVisualizer(
                dataset_name, output_folder, True, vist_cfg.launch.visualize
            )
            inference_on_dataset(model, data_loader, visualizer)


def train(cfg: Config) -> Dict[str, EvalResults]:
    """Training function."""
    det2cfg = to_detectron2(cfg)
    default_setup(cfg, det2cfg, cfg.launch)

    trainer = DefaultTrainer(cfg, det2cfg)
    trainer.cfg.MODEL.WEIGHTS = cfg.launch.weights
    trainer.resume_or_load(resume=cfg.launch.resume)
    return trainer.train()


def test(cfg: Config) -> Dict[str, EvalResults]:
    """Test function."""
    det2cfg = to_detectron2(cfg)
    default_setup(cfg, det2cfg, cfg.launch)

    model = build_model(cfg.model)
    model.to(torch.device(cfg.launch.device))
    Checkpointer(model, save_dir=det2cfg.OUTPUT_DIR).resume_or_load(
        cfg.launch.weights, resume=cfg.launch.resume
    )
    return DefaultTrainer.test_static(cfg, det2cfg, model)


def predict(cfg: Config) -> None:
    """Prediction function."""
    det2cfg = to_detectron2(cfg)
    default_setup(cfg, det2cfg, cfg.launch)

    model = build_model(cfg.model)
    model.to(torch.device(cfg.launch.device))
    Checkpointer(model, save_dir=det2cfg.OUTPUT_DIR).resume_or_load(
        cfg.launch.weights, resume=cfg.launch.resume
    )
    DefaultTrainer.predict(cfg, model)



"""VisT tracking loss implementations."""

from .base import BaseLoss, LossConfig, build_loss
from .embedding_distance import EmbeddingDistanceLoss
from .multi_pos_cross_entropy import MultiPosCrossEntropyLoss

__all__ = [
    "BaseLoss",
    "build_loss",
    "LossConfig",
    "EmbeddingDistanceLoss",
    "MultiPosCrossEntropyLoss",
]



"""VisT struct module."""
from .data import Extrinsics, Images, InputSample, Intrinsics
from .labels import Boxes2D, Boxes3D
from .structures import (
    DataInstance,
    DictStrAny,
    EvalResult,
    EvalResults,
    LabelInstance,
    LossesType,
    ModelOutput,
    NDArrayF64,
    NDArrayUI8,
    TorchCheckpoint,
)

__all__ = [
    "Boxes2D",
    "Boxes3D",
    "DataInstance",
    "LabelInstance",
    "TorchCheckpoint",
    "NDArrayF64",
    "NDArrayUI8",
    "LossesType",
    "EvalResult",
    "EvalResults",
    "InputSample",
    "Images",
    "Intrinsics",
    "Extrinsics",
    "ModelOutput",
    "DictStrAny",
]



"""Load and convert kitti labels to scalabel format."""
import inspect
import os
import os.path as osp
from typing import Optional

from scalabel.label.from_kitti import from_kitti
from scalabel.label.io import load, load_label_config, save
from scalabel.label.typing import Dataset

from .base import BaseDatasetConfig, BaseDatasetLoader


class KITTIDatasetConfig(BaseDatasetConfig):
    """Config for training/evaluation datasets."""

    split: Optional[str]
    data_type: Optional[str]


class KITTI(BaseDatasetLoader):  # pragma: no cover
    """KITTI dataloading class."""

    def __init__(self, cfg: BaseDatasetConfig):
        """Init dataset loader."""
        super().__init__(cfg)
        self.cfg = KITTIDatasetConfig(**cfg.dict())  # type: KITTIDatasetConfig

    def load_dataset(self) -> Dataset:
        """Convert kitti annotations to Scalabel format."""
        assert self.cfg.annotations is not None
        cfg_path = self.cfg.config_path
        if cfg_path is None:
            cfg_path = os.path.join(
                os.path.dirname(os.path.abspath(inspect.stack()[1][1])),
                "kitti.toml",
            )
        metadata_cfg = load_label_config(cfg_path)

        if not os.path.exists(self.cfg.annotations):
            assert self.cfg.data_type is not None
            assert self.cfg.split is not None
            data_dir = osp.join(
                self.cfg.data_root, self.cfg.data_type, self.cfg.split
            )
            frames = from_kitti(data_dir, self.cfg.data_type)
            save(self.cfg.annotations, frames)
        else:
            frames = load(
                self.cfg.annotations,
                validate_frames=self.cfg.validate_frames,
                nprocs=self.cfg.num_processes,
            ).frames

        return Dataset(frames=frames, config=metadata_cfg)



"""Config definitions."""
from .config import Config, Launch, Solver, parse_config
from .defaults import default_argument_parser

__all__ = [
    "Config",
    "Launch",
    "parse_config",
    "Solver",
    "default_argument_parser",
]
